"""
Manipulate stochastic processes
===============================
"""

# %%
# The objective here is to manipulate a multivariate stochastic process :math:`X: \Omega \times \mathcal{D} \rightarrow \mathbb{R}^d`,
# where :math:`\mathcal{D} \in \mathbb{R}^n` is discretized on the mesh :math:`\mathcal{M}`
# and exhibit some of the services exposed by the :class:`~openturns.Process` objects:
#
# - ask for the dimension, with the method :meth:`~openturns.Process.getOutputDimension` ;
# - ask for the mesh, with the method :meth:`~openturns.Process.getMesh` ;
# - ask for the mesh as regular 1-d mesh, with the :meth:`~openturns.Process.getTimeGrid` method ;
# - ask for a realization, with the method the :meth:`~openturns.Process.getRealization` method ;
# - ask for a continuous realization, with the :meth:`~openturns.Process.getContinuousRealization` method ;
# - ask for a sample of realizations, with the :meth:`~openturns.Process.getSample` method ;
# - ask for the normality of the process with the :meth:`~openturns.Process.isNormal` method ;
# - ask for the stationarity of the process with the :meth:`~openturns.Process.isStationary` method.
#

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We create a mesh -a time grid- which is a :class:`~openturns.RegularGrid` :
tMin = 0.0
timeStep = 0.1
n = 100
time_grid = ot.RegularGrid(tMin, timeStep, n)
time_grid.setName("time")

# %%
# We create a Normal process :math:`X_t = (X_t^0, X_t^1, X_t^2)` in dimension 3 with an exponential covariance model.
# We define the amplitude and the scale of the :class:`~openturns.ExponentialModel`
scale = [4.0]
amplitude = [1.0, 2.0, 3.0]

# %%
# We define a spatial correlation :
spatialCorrelation = ot.CorrelationMatrix(3)
spatialCorrelation[0, 1] = 0.8
spatialCorrelation[0, 2] = 0.6
spatialCorrelation[1, 2] = 0.1

# %%
# The covariance model is now created with :
myCovarianceModel = ot.ExponentialModel(scale, amplitude, spatialCorrelation)

# %%
# Eventually, the process is  built with :
process = ot.GaussianProcess(myCovarianceModel, time_grid)

# %%
# The dimension `d` of the process may be retrieved by
dim = process.getOutputDimension()
print("Dimension : %d" % dim)

# %%
# The underlying mesh of the process is obtained with the `getMesh` method :
mesh = process.getMesh()

# %%
# We have access to peculiar data of the mesh such as the corners :
minMesh = mesh.getVertices().getMin()[0]
maxMesh = mesh.getVertices().getMax()[0]

# %%
# We draw it :
graph = mesh.draw()
graph.setTitle("Time grid (mesh)")
graph.setXTitle("t")
graph.setYTitle("")
view = otv.View(graph)


# %%
# We can get the time grid of the process when the mesh can be interpreted as a regular time grid :
print(process.getTimeGrid())


# %%
# A typical feature of a process is the generation of a realization of itself :
realization = process.getRealization()

# %%
# Here it is a sample of size :math:`100 \times 4` (100 time steps, 3 spatial cooordinates and the time variable).
# We are able to draw its marginals, for instance the first (index 0) one :math:`X_t^0`, against the time with no interpolation:
interpolate = False
graph = realization.drawMarginal(0, interpolate)
graph.setTitle("First marginal of a realization of the process")
graph.setXTitle("t")
graph.setYTitle(r"$X_t^0$")
view = otv.View(graph)

# %%
# The same graph, but with interpolated values (default behaviour) :
graph = realization.drawMarginal(0)
graph.setTitle("First marginal of a realization of the process")
graph.setXTitle("t")
graph.setYTitle(r"$X_t^0$")
view = otv.View(graph)


# %%
# We can build a function representing the process using P1-Lagrange interpolation (when not defined from a functional model).
continuousRealization = process.getContinuousRealization()

# %%
# Once again we draw its first marginal :
marginal0 = continuousRealization.getMarginal(0)
graph = marginal0.draw(minMesh, maxMesh)
graph.setTitle("First marginal of a P1-Lagrange continuous realization of the process")
graph.setXTitle("t")
view = otv.View(graph)

# %%
# Please note that the `marginal0` object is a function. Consequently we can
# evaluate it at any point of the domain, say :math:`t_0=3.5678`
t0 = 3.5678
print(t0, marginal0([t0]))


# %%
# Several realizations of the process may be determined at once :
number = 10
fieldSample = process.getSample(number)

# %%
# Let us draw them the first marginal
# sphinx_gallery_thumbnail_number = 5
graph = fieldSample.drawMarginal(0, False)
graph.setTitle("First marginal of 10 realizations of the process")
graph.setXTitle("t")
graph.setYTitle(r"$X_t^0$")
view = otv.View(graph)

# %%
# Same graph, but with interpolated values :
graph = fieldSample.drawMarginal(0)
graph.setTitle("First marginal of 10 realizations of the process")
graph.setXTitle("t")
graph.setYTitle(r"$X_t^0$")
view = otv.View(graph)


# %%
# Miscellanies
# ------------
#
# We can extract any marginal of the process with the `getMarginal` method.
# Beware the numerotation begins at 0 ! It may be not implemented yet for
# some processes.
# The extracted marginal is a 1-d Gaussian process
print(process.getMarginal([1]))

# %%
# If we extract simultaneously two indices we build a 2-d Gaussian process
print(process.getMarginal([0, 2]))

# %%
# We can check whether the process is Gaussian or not
print("Is normal ? ", process.isNormal())

# %%
# and the stationarity as well
print("Is stationary ? ", process.isStationary())

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
from openturns.usecases import flood_model

ot.TESTPREAMBLE()

ot.Log.Show(ot.Log.NONE)

sampleSize = 500
totalDegree = 7

#
print("+ Compute flood model using least squares")
fm = flood_model.FloodModel()
inputDescription = fm.model.getInputDescription()
marginals = [fm.distribution.getMarginal(i) for i in range(fm.dim)]
basis = ot.OrthogonalProductPolynomialFactory(marginals)
inputSample = fm.distribution.getSample(sampleSize)
outputSample = fm.model(inputSample)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
leastSquaresStrategy = ot.LeastSquaresStrategy(selectionAlgorithm)
print("leastSquaresStrategy")
print(leastSquaresStrategy)
print("leastSquaresStrategy (repr)")
print(leastSquaresStrategy.__repr__())
print("leastSquaresStrategy (html)")
print(leastSquaresStrategy._repr_html_())
assert leastSquaresStrategy.isLeastSquares()
assert leastSquaresStrategy.involvesModelSelection()
assert selectionAlgorithm.involvesModelSelection()

enumerateFunction = basis.getEnumerateFunction()
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
adaptiveStrategy = ot.FixedStrategy(basis, basisSize)
chaosAlgo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, fm.distribution, adaptiveStrategy, leastSquaresStrategy
)
projectionStrategy = chaosAlgo.getProjectionStrategy()
print("projectionStrategy")
print(projectionStrategy)
print("projectionStrategy (repr)")
print(projectionStrategy.__repr__())
print("projectionStrategy (html)")
print(projectionStrategy._repr_html_())
assert not adaptiveStrategy.involvesModelSelection()


#
print("+ Compute flood model by integration")
integrationStrategy = ot.IntegrationStrategy()
print("integrationStrategy")
print(integrationStrategy)
print("integrationStrategy (repr)")
print(integrationStrategy.__repr__())
print("integrationStrategy (html)")
print(integrationStrategy._repr_html_())
chaosAlgo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, fm.distribution, adaptiveStrategy, integrationStrategy
)
projectionStrategy = chaosAlgo.getProjectionStrategy()
print("projectionStrategy")
print(projectionStrategy)
print("projectionStrategy (repr)")
print(projectionStrategy.__repr__())
print("projectionStrategy (html)")
print(projectionStrategy._repr_html_())
assert not integrationStrategy.isLeastSquares()

#
print("+ Compute flood model with large output dimension")
outputDimension = 20
fm = flood_model.FloodModel()
model = ot.AggregatedFunction([fm.model] * outputDimension)
sampleSize = 500
totalDegree = 7
fm = flood_model.FloodModel()
inputDescription = fm.model.getInputDescription()
marginals = [fm.distribution.getMarginal(i) for i in range(fm.dim)]
basis = ot.OrthogonalProductPolynomialFactory(marginals)
inputSample = fm.distribution.getSample(sampleSize)
outputSample = model(inputSample)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
leastSquaresStrategy = ot.LeastSquaresStrategy(selectionAlgorithm)
print("leastSquaresStrategy")
print(leastSquaresStrategy)
print("leastSquaresStrategy (repr)")
print(leastSquaresStrategy.__repr__())
print(leastSquaresStrategy._repr_html_())

enumerateFunction = basis.getEnumerateFunction()
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
adaptiveStrategy = ot.FixedStrategy(basis, basisSize)
chaosAlgo = ot.FunctionalChaosAlgorithm(
    inputSample, outputSample, fm.distribution, adaptiveStrategy, leastSquaresStrategy
)
projectionStrategy = chaosAlgo.getProjectionStrategy()
print("projectionStrategy")
print(projectionStrategy)
print("projectionStrategy (repr)")
print(projectionStrategy.__repr__())
print("projectionStrategy (html)")
print(projectionStrategy._repr_html_())



"""
Use case : the Chaboche mechanical model
========================================
"""

import openturns as ot


class ChabocheModel:
    r"""
    Data class for the Chaboche mechanical model.

    Parameters
    ----------
    strainMin : float, optional
        The minimum value of the strain. The default is 0.0.

    strainMax : float, optional
        The maximum value of the strain. The default is 0.07

    trueR : float, optional
        The true value of the R parameter. The default is 750.0e6.

    trueC : float, optional
        The true value of the C parameter. The default is 2750.0e6.

    trueGamma : float, optional
        The true value of the Gamma parameter. The default is 10.0.


    Attributes
    ----------
    dim : int
        Dimension of the problem, dim=4.

    Strain : :class:`~openturns.Uniform` distribution
        `Uniform(strainMin, strainMax)`

    R : :class:`~openturns.LogNormal` distribution
        `LogNormal().setParameter(ot.LogNormalMuSigma()([750.0e6, 11.0e6, 0.0]))`

    C : :class:`~openturns.Normal` distribution
        `Normal(2750.0e6, 250.0e6)`

    Gamma : :class:`~openturns.Normal` distribution
        `Normal(10.0, 2.0)`

    distribution : :class:`~openturns.JointDistribution`
        The joint distribution of the input parameters.

    model : :class:`~openturns.Function`
        The Chaboche mechanical law.
        The model has input dimension 4 and output dimension 1.
        More precisely, we have :math:`\vect{X} = (\epsilon, R,
        C, \gamma)` and  :math:`Y = \sigma`.

    data : :class:`~openturns.Sample`
        A data set of size 10 and dimension 2 which contains noisy observations of the strain (column 0)
        and the stress (column 1).

    Examples
    --------
    >>> from openturns.usecases import chaboche_model
    >>> # Load the Chaboche model
    >>> cm = chaboche_model.ChabocheModel()
    >>> print(cm.data[:5])
            [ Strain      Stress (Pa) ]
    0 : [ 0           7.56e+08    ]
    1 : [ 0.0077      7.57e+08    ]
    2 : [ 0.0155      7.85e+08    ]
    3 : [ 0.0233      8.19e+08    ]
    4 : [ 0.0311      8.01e+08    ]
    >>> print("Inputs:", cm.model.getInputDescription())
    Inputs: [Strain,R,C,Gamma]
    >>> print("Outputs:", cm.model.getOutputDescription())
    Outputs: [Sigma]
    """

    def __init__(
        self,
        strainMin=0.0,
        strainMax=0.07,
        trueR=750.0e6,
        trueC=2750.0e6,
        trueGamma=10.0,
    ):
        self.dim = 4  # number of inputs

        if strainMin >= strainMax:
            raise ValueError(
                f"Strain minimum = {strainMin} >= strainMax = {strainMax}."
            )
        self.strainMin = strainMin
        self.strainMax = strainMax
        self.Strain = ot.Uniform(strainMin, strainMax)
        self.Strain.setDescription(["Strain"])

        self.trueR = trueR
        self.R = ot.LogNormal()
        self.R.setParameter(ot.LogNormalMuSigma()([750.0e6, 11.0e6, 0.0]))
        self.R.setDescription(["R"])

        self.trueC = trueC
        self.C = ot.Normal(2750.0e6, 250.0e6)
        self.C.setDescription(["C"])

        self.trueGamma = trueGamma
        self.Gamma = ot.Normal(10.0, 2.0)
        self.Gamma.setDescription(["Gamma"])

        self.distribution = ot.JointDistribution(
            [self.Strain, self.R, self.C, self.Gamma]
        )
        self.inputDistribution = self.distribution  # deprecated
        self.model = ot.SymbolicFunction(
            ["Strain", "R", "C", "Gamma"], ["R - C * expm1(-Gamma * Strain) / Gamma"]
        )
        self.model.setOutputDescription(["Stress"])

        self.data = ot.Sample(
            [
                [0.0, 7.56e08],
                [0.0077, 7.57e08],
                [0.0155, 7.85e08],
                [0.0233, 8.19e08],
                [0.0311, 8.01e08],
                [0.0388, 8.42e08],
                [0.0466, 8.49e08],
                [0.0544, 8.79e08],
                [0.0622, 8.85e08],
                [0.07, 8.96e08],
            ]
        )
        self.data.setDescription(["Strain", "Stress (Pa)"])



"""
Calibrate a parametric model: a quick-start guide to calibration
================================================================

In this example we present the calibration of a parametric module.
To do this, we show how to define the observations.
Then we create a Python function and create the parametric model that is
to be calibrated.
Finally, we calibrate the parameters of the model using least squares
and we validate the hypotheses of the method.

Please read :ref:`code_calibration` for more details on code calibration and least squares.
In this example we are interested in the calibration of the :ref:`flooding model <use-case-flood-model>`.
Once the reader has mastered this example, the :doc:`calibration of the Chaboche mechanical model
</auto_calibration/least_squares_and_gaussian_calibration/plot_calibration_chaboche>` may be
considered to make an in-depth study of these algorithms.
The goal of this script is to be relatively small, so please consider reading
the other examples if this is relevant.
"""

# %%
# Parameters to calibrate and observations
# ----------------------------------------
#
# The parametric model is:
#
# .. math::
#    H = a Q^b.
#
# The vector of parameters to calibrate is:
#
# .. math::
#    \theta = (a,b).
#
# We consider a sample size equal to:
#
# .. math::
#    n = 10.
#
# The observations are the couples :math:`\{(Q_i,H_i)\}_{i=1,...,n}`, i.e. each observation is a
# couple made of the flowrate and the corresponding river height.
#
# We choose to calibrate this model using non linear least squares, because
# this is a relatively flexible method.

# %%
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import flood_model

# %%
# Define the observations
# -----------------------
# In practice, we generally use a data set which has been obtained from
# measurements.
# This data set can be loaded using e.g. :meth:`~openturns.Sample.ImportFromCSVFile`.
# Here we import the data from the
# :class:`~openturns.usecases.flood_model.FloodModel`
# class.

# %%
fm = flood_model.FloodModel()
Qobs = fm.data[:, 0]
Hobs = fm.data[:, 1]
nbobs = fm.data.getSize()
fm.data


# %%
# Create the physical model
# -------------------------
# We define the model :math:`g` which has 3 inputs and one output `H`.
# This model has a parametric shape that may correspond to the data,
# which has some power shape.
# In the model, the parameters are `a` and `b`, the input is `Q`
# and the output is `H`:
# - `Q` : the flowrate of the river,
# - `a`, `b` : the parameters.


# %%
def functionSimpleFlooding(X):
    Q, a, b = X
    H = a * Q**b
    return [H]


g = ot.PythonFunction(3, 1, functionSimpleFlooding)
g = ot.MemoizeFunction(g)
g.setInputDescription(["Q", "a", "b"])
g.setOutputDescription(["H"])


# %%
# Setting the calibration parameters
# ----------------------------------

# %%
# Define the value of the reference values of the :math:`\theta` parameter.
# There is no particular method to set these values: we used
# trial-and-error to see the order of magnitude of the parameters.

# %%
aInitial = 0.1
bInitial = 0.5
thetaPrior = [aInitial, bInitial]

# %%
# Create the parametric function
# ------------------------------
# In the physical model, the inputs and parameters are ordered as
# presented in the next table.
# Notice that there are no parameters in the physical model.
#
# +-------+----------------+
# | Index | Input variable |
# +=======+================+
# | 0     | Q              |
# +-------+----------------+
# | 1     | a              |
# +-------+----------------+
# | 2     | b              |
# +-------+----------------+
#
# +-------+-----------+
# | Index | Parameter |
# +=======+===========+
# | ∅     | ∅         |
# +-------+-----------+
#
# **Table 1.** Indices and names of the inputs and parameters of the physical model.
#

# %%
print("Physical Model Inputs:", g.getInputDescription())
print("Physical Model Parameters:", g.getParameterDescription())

# %%
# In order to perform calibration, we have to define a parametric model,
# with observed inputs and parameters to calibrate.
# In order to do this, we create a :class:`~openturns.ParametricFunction`
# where the parameters are `a` and `b` which have the indices 1 and
# and 2 in the physical model.
# Please read :doc:`Create a parametric function </auto_functional_modeling/vectorial_functions/plot_parametric_function>`
# for more details on this topic.
#
# +-------+----------------+
# | Index | Input variable |
# +=======+================+
# | 0     | Q              |
# +-------+----------------+
#
# +-------+-----------+
# | Index | Parameter |
# +=======+===========+
# | 0     | a         |
# +-------+-----------+
# | 1     | b         |
# +-------+-----------+
#
# **Table 2.** Indices and names of the inputs and parameters of the parametric model.
#


# %%
# The following statement create the calibrated function from the model.
# The calibrated parameters :math:`a`, :math:`b` are at
# indices 1, 2 in the inputs arguments of the model.

# %%
calibratedIndices = [1, 2]
mycf = ot.ParametricFunction(g, calibratedIndices, thetaPrior)

# %%
# Plot the Y observations versus the X observations.

# %%
title = "Before calibration : a = %.4f, b = %.4f" % (aInitial, bInitial)
graph = ot.Graph(title, "Q", "H")
# Plot the model before calibration
curve = mycf.draw(100.0, 4000.0).getDrawable(0)
curve.setLegend("Model, before calibration")
graph.add(curve)
# Plot the noisy observations
cloud = ot.Cloud(Qobs, Hobs)
cloud.setLegend("Observations")
graph.add(cloud)
#
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# Wee see that the model does not fit to the data.
# The goal of calibration is to find which parameter best fit to the
# observations.

# %%
# Calibration with non linear least squares
# -----------------------------------------
# The :class:`~openturns.NonLinearLeastSquaresCalibration` class performs
# non linear least squares calibration by minimizing the squared Euclidian norm
# between the predictions and the observations.

# %%
algo = ot.NonLinearLeastSquaresCalibration(mycf, Qobs, Hobs, thetaPrior)

# %%
# The :meth:`~openturns.NonLinearLeastSquaresCalibration.run` method computes
# the solution of the problem.

# %%
algo.run()
calibrationResult = algo.getResult()

# %%
# Analysis of the results
# -----------------------

# %%
# The :meth:`~openturns.CalibrationResult.getParameterMAP` method returns the
# maximum of the posterior distribution of :math:`\theta`.

# %%
thetaMAP = calibrationResult.getParameterMAP()
print(thetaMAP)


# %%
# In order to see if the parameters fit the data,
# we plot the input depending on the output before and after calibration.

# sphinx_gallery_thumbnail_number = 2
graph = calibrationResult.drawObservationsVsInputs()
aEstimated, bEstimated = thetaMAP
title = "After calibration : a = %.4f, b = %.4f" % (aEstimated, bEstimated)
graph.setTitle(title)
graph.setLegendPosition("upper left")
view = otv.View(graph)


# %%
# One of the hypotheses of the least squares method is that
# the residuals follow a normal distribution: the next
# cell checks if this hypothesis is satisfied here.

# %%
graph = calibrationResult.drawResiduals()
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# The analysis of the residuals shows that the distribution is centered on
# zero and symmetric.
# This indicates that the calibration performed well.
# Moreover, the distribution of the residuals is close to being Gaussian.

# %%
# *Remark.* The logarithm of the height is:
#
# .. math::
#    \log(H) = \log(a) + b \log(Q).
#
# Hence, transforming the data on a logarithmic scale leads to a parametric
# model that is linear in the parameters.
# The parameters of this transformed model can be estimated using linear
# linear squares, which may lead to a significant improvement in terms
# of number of function evaluations.
#

otv.View.ShowAll()



import openturns as ot
import openturns.viewer as otv

generator = ot.Normal(2)
size = 5
array = []
palette = ot.Drawable.BuildDefaultPalette(size)
for i in range(size):
    vertices = generator.getSample(3)
    array.append(ot.Polygon(vertices, palette[i], palette[size - i - 1]))

graph = ot.Graph("PolygonArray example", "x1", "x2")
graph.add(ot.PolygonArray(array))

otv.View(graph, figure_kw={"figsize": (4, 4)})



import openturns as ot
import openturns.viewer as otv


size = 100
normal = ot.Normal(1)
sample = normal.getSample(size)

henryPlot = ot.VisualTest.DrawHenryLine(sample)

otv.View(henryPlot, figure_kw={"figsize": (4.5, 4.5)})



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
distribution = otexp.MarginalUniformOrderStatistics(4, [0, 1, 3])
print("Distribution ", repr(distribution))
print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", oneRealization)

parameters = distribution.getParametersCollection()
print("parameters=", parameters)
print("Standard representative=", distribution.getStandardRepresentative())

marginal = distribution.getMarginal([0, 2])
print("Marginal [0, 2]=", marginal)

validation = ott.DistributionValidation(distribution)
validation.skipDependenceMeasures()  # too slow
validation.run()



"""
Create a general linear model metamodel
=======================================
"""

# %%
# In this example we create a global approximation of a model response using a general linear model.
# We show how to use the :class:`~openturns.GaussianProcessFitter` class, which estimates the parameters of the model.

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# We create a model and a sample from this model.

# %%
dimension = 2
input_names = ["x1", "x2"]
formulas = ["cos(0.1 * x1 + 0.2 * x2)"]
model = ot.SymbolicFunction(input_names, formulas)
distribution = ot.Normal(dimension)
x = distribution.getSample(100)
y = model(x)

# %%
# We create a :class:`~openturns.GaussianProcessFitter` based on a quadratic basis.
# The `run` method estimates the coefficients of the trend and the hyperparameters of the covariance model.

# %%
basis = ot.QuadraticBasisFactory(dimension).build()
covarianceModel = ot.SquaredExponential([1] * dimension, [1.0])
algo = ot.GaussianProcessFitter(x, y, covarianceModel, basis)
algo.run()
result = algo.getResult()

# %%
# We see that the trend coefficients have been estimated.

# %%
result.getTrendCoefficients()

# %%
# The parameters of the covariance models also have been estimated.

# %%
result.getCovarianceModel()

# %%
# The `getMetaModel` method returns the metamodel where the parameters have been estimated.

# %%
responseSurface = result.getMetaModel()

# %%
# Plot the output of our model depending on :math:`x_2` with :math:`x_1=0.5`.

# %%
xmin = -5.0
xmax = 5.0
x1value = 0.5
parametricModelGraph = ot.ParametricFunction(model, [0], [x1value]).draw(xmin, xmax)
graphMetamodel = ot.ParametricFunction(responseSurface, [0], [x1value]).draw(xmin, xmax)
parametricModelGraph.add(graphMetamodel)
parametricModelGraph.setLegends(["Model", "Meta-Model"])
parametricModelGraph.setLegendPosition("upper right")
view = otv.View(parametricModelGraph)

# %%
view.ShowAll()



import openturns as ot
import openturns.viewer as otv

ot.ResourceMap.SetAsUnsignedInteger("Distribution-DefaultIntegrationNodesNumber", 64)
ot.ResourceMap.SetAsUnsignedInteger("Mesh-LargeSize", 0)

grid = ot.GridLayout(2, 2)
threshold = 0.2
inVars = ["x1", "x2"]
formula = "sin(x1)*sin(x2)"
f = ot.SymbolicFunction(inVars, [formula])
levelSet = ot.LevelSet(f, ot.Less(), threshold)
xMin = [-5.0] * 2
xMax = [5.0] * 2
box = ot.Interval(xMin, xMax)
mesh = ot.LevelSetMesher([51] * 2).build(levelSet, box, False)
pdf_2d = ot.Graph("Normal distribution over a mesh, PDF", "x1", "x2")
cdf_2d = ot.Graph("Normal distribution over a mesh, CDF", "x1", "x2")
dist = ot.Normal([0] * 2, [2] * 2)
distribution_2d = ot.TruncatedOverMesh(dist, mesh)
mesh_g = mesh.draw()
mesh_g.setLegends([""])
mesh_g.setTitle("Mesh of domain D={" + formula + "<=" + str(threshold) + "}")
mesh_g.setXTitle("x1")
mesh_g.setYTitle("x2")

iso = f.draw(xMin, xMax)
iso.setTitle("Normal sample on D")
dr = iso.getDrawable(0)
dr.setLevels([threshold])
iso.setDrawables([dr])
cloud = ot.Cloud(distribution_2d.getSample(1500))
cloud.setColor("blue")
iso.add(cloud)
grid.setGraph(0, 0, mesh_g)
grid.setGraph(0, 1, iso)

pdf_2d.add(distribution_2d.drawPDF(xMin, xMax))
pdf_2d.setLegends([""])
cdf_2d.add(distribution_2d.drawCDF(xMin, xMax, [41] * 2))
cdf_2d.setLegends([""])
grid.setGraph(1, 0, pdf_2d)
grid.setGraph(1, 1, cdf_2d)
grid.setTitle("TruncatedOverMesh")
grid.setLegendPosition("upper right")
v = otv.View(grid)
fig = v.getFigure()
fig.axes[1].legend(loc="best")



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

dimension = 2
sampleSize = 100000

# we create an analytical function
analytical = ot.SymbolicFunction(["x0", "x1"], ["10+3*x0+x1"])

# we create a collection of centered and reduced gaussian distributions
aCollection = [ot.Normal()] * dimension

# we create one distribution object
aDistribution = ot.JointDistribution(aCollection)

# Random vectors
randomVector = ot.RandomVector(aDistribution)

# we create two input samples for the function
inputSample = randomVector.getSample(sampleSize)
outputSample = analytical(inputSample)

# Create the CorrelationAnalysis object
corr_analysis = ot.CorrelationAnalysis(inputSample, outputSample)

# Because here outputSample = analyical(inputSample),
# the theoretical squared SRC indices are equal to the Sobol indices
# See theory/reliability_sensitivity/sensitivity_sobol.rst for the computation.
squared_src = corr_analysis.computeSquaredSRC()
ott.assert_almost_equal(squared_src, [0.9, 0.1], 0.0, 1e-2)  # theoretical value

# Squared SRC with normalize
squared_src_normalize = corr_analysis.computeSquaredSRC(True)
ott.assert_almost_equal(
    squared_src_normalize, [0.9, 0.1], 0.0, 1e-2
)  # theoretical value

src = corr_analysis.computeSRC()
ott.assert_almost_equal(
    src, [0.9486832980505138, 0.31622776601683794], 0.0, 1e-2
)  # sqrt of squared_src

srrc = corr_analysis.computeSRRC()
ott.assert_almost_equal(srrc, [0.94, 0.30], 0.0, 1e-2)  # approximate value

pcc = corr_analysis.computePCC()
ott.assert_almost_equal(pcc, [1.0, 1.0], 1e-5, 0.0)  # theoretical value

prcc = corr_analysis.computePRCC()
ott.assert_almost_equal(prcc, [0.99, 0.92], 0.0, 1e-2)  # approximate value

pearson = corr_analysis.computeLinearCorrelation()
ott.assert_almost_equal(pearson, [0.95, 0.31], 0.0, 1e-2)  # approximate value

spearman = corr_analysis.computeSpearmanCorrelation()
ott.assert_almost_equal(spearman, [0.94, 0.30], 0.0, 1e-2)  # approximate value

kendalltau = corr_analysis.computeKendallTau()
ott.assert_almost_equal(kendalltau, [0.79, 0.20], 0.0, 1e-2)  # approximate value



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott

ot.TESTPREAMBLE()

f1 = ot.SymbolicFunction(
    ["a0", "a1"],
    ["-4 * exp((-25 / 8) * (a0^2 + a1^2)) + 7 * exp((-125 / 4) * (a0^2 + a1^2))"] * 2,
)

distribution = ot.JointDistribution([ot.Uniform(-1.0, 1.0)] * 2)
x0 = ot.LowDiscrepancyExperiment(ot.HaltonSequence(), distribution, 100).generate()
y0 = f1(x0)

# reference mean for each generated batch of input x
ref_mean = [
    [-0.0271247, -0.0171091],
    [-0.072897, 0.0526871],
    [-0.0322869, -0.0448992],
    [0.0364691, 0.0630369],
    [-0.0838352, -0.0595057],
]

algo = otexp.LOLAVoronoi(x0, y0, distribution)
newX = ot.Sample(0, x0.getDimension())
inc = 15
for i in range(5):
    x = algo.generate(inc)
    newX.add(x)
    y = f1(x)
    algo.update(x, y)

    x_mean = x.computeMean()
    print(f"iteration={i} x_mean={x_mean}")

    lolaScore = algo.getLOLAScore()
    voronoiScore = algo.getVoronoiScore()
    assert len(lolaScore) == len(x0) + i * inc
    assert len(lolaScore) == len(voronoiScore)

    # FIXME: legacy KDTree is incorrect https://github.com/openturns/openturns/issues/2617
    if ot.PlatformInfo.HasFeature("nanoflann"):
        ott.assert_almost_equal(x_mean, ref_mean[i])

    if False:
        import openturns.viewer as otv

        cloud1 = ot.Cloud(x0)
        cloud1.setPointStyle("fcircle")
        cloud1.setColor("blue")
        cloud2 = ot.Cloud(newX)
        cloud2.setPointStyle("fcircle")
        cloud2.setColor("red")
        graph = ot.Graph("LOLA-Voronoi", "x1", "x2")
        graph.add(cloud1)
        graph.add(cloud2)
        otv.View(graph)
        otv.View.ShowAll()

# linear model: the non linearity score should have no effect
ot.ResourceMap.SetAsString("LOLAVoronoi-NonLinearityAggregationMethod", "Maximum")
f2 = ot.SymbolicFunction(["x0", "x1"], ["2 * x0 + 3 * x1 + 8"])
y0 = f2(x0)
algo = otexp.LOLAVoronoi(x0, y0, distribution)
for i in range(10):
    x = algo.generate(20)
    y = f2(x)
    algo.update(x, y)
    x_mean = x.computeMean()
    print(f"iteration={i} x_mean={x_mean}")
ott.assert_almost_equal(algo.getLOLAScore().asPoint().norm(), 0.0)
X = algo.getInputSample()
discrepancy = ot.SobolSequence.ComputeStarDiscrepancy(X)
assert discrepancy < 1.0, "wrong repartition"

# oscillating model on a grid
ot.ResourceMap.SetAsUnsignedInteger("LOLAVoronoi-MaximumCombinationsNumber", 3870)
f3 = ot.SymbolicFunction(["x0", "x1"], ["sin(10 * x0) + cos(10 * x1)"])
x0 = ot.Box([8, 8]).generate()
y0 = f3(x0)
algo = otexp.LOLAVoronoi(x0, y0, distribution)
for i in range(10):
    x = algo.generate(20)
    y = f3(x)
    algo.update(x, y)
    x_mean = x.computeMean()
    print(f"iteration={i} x_mean={x_mean}")
X = algo.getInputSample()
discrepancy = ot.SobolSequence.ComputeStarDiscrepancy(X)
assert discrepancy < 1.0, "wrong repartition"



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Multivariate case
coll2 = ot.DistributionCollection(0)
coll2.add(ot.Dirac(1))
coll2.add(ot.Dirac(2))
coll2.add(ot.Bernoulli(0.7))
coll2.add(ot.Uniform(3.0, 4.0))
d2 = ot.JointDistribution(coll2)
coll1 = ot.DistributionCollection(0)
coll1.add(ot.Uniform())
coll1.add(ot.Uniform())
d1 = ot.JointDistribution(coll1)
# Test the different DOE
ot.ResourceMap.SetAsUnsignedInteger(
    "CompoundDistribution-MarginalIntegrationNodesNumber", 32
)
ot.ResourceMap.SetAsUnsignedInteger(
    "CompoundDistribution-MaximumIntegrationNodesNumber", 10000
)
for method in ["GaussProduct", "QMC", "MC"]:
    print("#" * 50)
    print("method=", method)
    ot.ResourceMap.SetAsString(
        "CompoundDistribution-ContinuousDiscretizationMethod", method
    )
    distribution = ot.CompoundDistribution(d1, d2)
    dim = distribution.getDimension()
    print("distribution=", distribution)
    print("Parameters ", distribution.getParametersCollection())
    print("Mean ", distribution.getMean())
    cov = distribution.getCovariance()
    cov_ref = ot.CovarianceMatrix([[0.0833333, 0.0], [0.0, 0.751111]])
    ott.assert_almost_equal(cov, cov_ref, 1e-3, 1e-3)
    # Is this distribution an elliptical distribution?
    print("Elliptical distribution= ", distribution.isElliptical())
    # Has this distribution an elliptical copula?
    print("Elliptical copula= ", distribution.hasEllipticalCopula())
    # Has this distribution an independent copula?
    print("Independent copula= ", distribution.hasIndependentCopula())
    # Test for realization of distribution
    oneRealization = distribution.getRealization()
    print("oneRealization=", oneRealization)
    # Test for sampling
    size = 10
    oneSample = distribution.getSample(size)
    print("oneSample=", oneSample)

    # Test for sampling
    size = 10000
    anotherSample = distribution.getSample(size)
    print("anotherSample mean=", anotherSample.computeMean())
    print("anotherSample covariance=", anotherSample.computeCovariance())

    # Define a point
    zero = ot.Point(dim, 0.0)

    # Show PDF and CDF of zero point
    zeroPDF = distribution.computePDF(zero)
    zeroCDF = distribution.computeCDF(zero)
    print("Zero point= ", zero, " pdf=", zeroPDF, " cdf=", zeroCDF)

    # Get 95% quantile
    quantile = distribution.computeQuantile(0.95)
    print("Quantile=", quantile)
    print("CDF(quantile)= %.5g" % distribution.computeCDF(quantile))
    # Get 95% survival function
    inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
    print("InverseSurvival=", repr(inverseSurvival))
    print(
        "Survival(inverseSurvival)=%.6f"
        % distribution.computeSurvivalFunction(inverseSurvival)
    )

    # Confidence regions
    # interval, threshold = distribution.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
    # print("Minimum volume interval=", interval)
    # print("threshold=", Point(1, threshold))
    # levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
    # print("Minimum volume level set=", levelSet)
    # print("beta=", Point(1, beta))
    # interval, beta = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
    # print("Bilateral confidence interval=", interval)
    # print("beta=", Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, False
    )
    print("Unilateral confidence interval (lower tail)=", interval)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, True
    )
    print("Unilateral confidence interval (upper tail)=", interval)
    print("beta=", ot.Point(1, beta))

# "
# 1D tests
# Create a collection of distribution
conditionedDistribution = ot.Normal()
conditioningDistributionCollection = ot.DistributionCollection(0)
# First conditioning distribution: continuous/continuous
atoms = ot.DistributionCollection(0)
atoms.add(ot.Uniform(0.0, 1.0))
atoms.add(ot.Uniform(1.0, 2.0))
conditioningDistributionCollection.add(ot.JointDistribution(atoms))
# Second conditioning distribution: discrete/continuous
atoms = ot.DistributionCollection(0)
atoms.add(ot.Binomial(3, 0.5))
atoms.add(ot.Uniform(1.0, 2.0))
conditioningDistributionCollection.add(ot.JointDistribution(atoms))
# Third conditioning distribution: dirac/continuous
atoms = ot.DistributionCollection(0)
atoms.add(ot.Dirac(0.5))
atoms.add(ot.Uniform(1.0, 2.0))
conditioningDistributionCollection.add(ot.JointDistribution(atoms))
for i in range(conditioningDistributionCollection.getSize()):
    print("conditioning distribution=", conditioningDistributionCollection[i])
    distribution = ot.CompoundDistribution(
        conditionedDistribution, conditioningDistributionCollection[i]
    )
    dim = distribution.getDimension()
    print("Distribution ", distribution)
    print("Parameters ", distribution.getParametersCollection())
    print("Mean ", distribution.getMean())
    print("Covariance ", distribution.getCovariance())
    # Is this distribution an elliptical distribution?
    print("Elliptical distribution= ", distribution.isElliptical())

    # Has this distribution an elliptical copula?
    print("Elliptical copula= ", distribution.hasEllipticalCopula())

    # Has this distribution an independent copula?
    print("Independent copula= ", distribution.hasIndependentCopula())

    # Test for realization of distribution
    oneRealization = distribution.getRealization()
    print("oneRealization=", oneRealization)

    # Test for sampling
    size = 10
    oneSample = distribution.getSample(size)
    print("oneSample=", oneSample)

    # Test for sampling
    size = 10000
    anotherSample = distribution.getSample(size)
    print("anotherSample mean=", anotherSample.computeMean())
    print("anotherSample covariance=", anotherSample.computeCovariance())

    # Define a point
    zero = ot.Point(dim, 0.0)

    # Show PDF and CDF of zero point
    zeroPDF = distribution.computePDF(zero)
    zeroCDF = distribution.computeCDF(zero)
    print("Zero point= ", zero, " pdf=%.6f" % zeroPDF, " cdf=%.6f" % zeroCDF)

    # Get 95% quantile
    quantile = distribution.computeQuantile(0.95)
    print("Quantile=", quantile)
    print("CDF(quantile)= %.12g" % distribution.computeCDF(quantile))
    # Get 95% survival function
    inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
    print("InverseSurvival=", repr(inverseSurvival))
    print(
        "Survival(inverseSurvival)=%.6f"
        % distribution.computeSurvivalFunction(inverseSurvival)
    )



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)
ot.RandomGenerator.SetSeed(0)

# We create a Sample
point1 = [10.0, 20.0]
point2 = [11.0, 21.0]
point3 = [12.0, 22.0]
sample1 = ot.Sample([point1, point2, point3])
referencemin = point1
referencemax = point3

# Iterative extrema, one point at a time
dimension = 2
iterPoint = ot.IterativeExtrema(dimension)
iterPoint.increment(point1)
iterPoint.increment(point2)
iterPoint.increment(point3)
computedmin = iterPoint.getMin()
computedmax = iterPoint.getMax()
ott.assert_almost_equal(referencemin, computedmin)
ott.assert_almost_equal(referencemax, computedmax)
iteration = iterPoint.getIterationNumber()
ott.assert_almost_equal(iteration, 3)

# Iterative extrema, one single sample
iterSample = ot.IterativeExtrema(dimension)
iterSample.increment(sample1)
computedmin = iterSample.getMin()
computedmax = iterSample.getMax()
ott.assert_almost_equal(referencemin, computedmin)
ott.assert_almost_equal(referencemax, computedmax)
iteration = iterSample.getIterationNumber()
ott.assert_almost_equal(iteration, 3)

# Iterative extrema, one single sample, then one point at a time
iterMixed = ot.IterativeExtrema(dimension)
iterMixed.increment(sample1)
iterMixed.increment(point1)
iterMixed.increment(point2)
iterMixed.increment(point3)
computedmin = iterMixed.getMin()
computedmax = iterMixed.getMax()
ott.assert_almost_equal(referencemin, computedmin)
ott.assert_almost_equal(referencemax, computedmax)
iteration = iterMixed.getIterationNumber()
ott.assert_almost_equal(iteration, 6)



import sympy as sp

x, a, b, a_, b_, point, gamma_, expX, beta_, alpha_, mu, sigma, skew = sp.symbols(
    "x a b a_ b_ point gamma_ expX beta_ alpha_ mu sigma skew", real=True
)

alpha_ = sp.Symbol("alpha_", positive=True)

# skew=2(1+a)/(a-3)sqrt((a-2)/a)=>a(a-3)^2*s^2=4(1+a)^2(a-2)
fskew = sp.expand(
    alpha_ * (alpha_ - 3) ** 2 * skew**2 - 4 * (1 + alpha_) ** 2 * (alpha_ - 2)
)
# iskew = sp.solve(fskew, alpha_)

# iskew = sp.solveset(fskew, alpha_, domain=S.Reals)
# print(iskew)
# print(str(iskew).replace('skew**2', 'sk2').replace('skew**4', 'sk4').replace('skew**6', 'sk6'))

roots = sp.solve(
    2 * (alpha_ + 1) / (alpha_ - 3) * sp.sqrt((alpha_ - 2) / alpha_) - skew, alpha_
)
root1 = roots[0]


# roots = sp.solve(fskew, alpha_)

for root in roots:
    print(root)
    # print(str(root).replace('skew**2', 'sk2').replace('skew**4', 'sk4').replace('skew**6', 'sk6'))
    print(root.subs(skew, 5.4))
    fskew.subs(alpha_, root).subs(skew, 5.4)
    # print('-'*100)

root1 = roots[0]
print("alpha=", root1)
# print(sp.simplify(root1))
print(sp.simplify(fskew.subs(alpha_, root1)))


# print(root1.subs(skew, 5.4))



"""
Use the Kolmogorov/Lilliefors test
==================================
"""

# %%
# In this example we are going to perform a Kolmogorov or a Lilliefors goodness-of-fit test for a 1-d continuous distribution.

# %%
import openturns as ot


# %%
# Create the data.

# %%
distribution = ot.Normal()
sample = distribution.getSample(50)

# %%
# Case 1 : the distribution parameters are known.
# -----------------------------------------------
#
# In the case where the parameters of the distribution are known,
# we must use the `Kolmogorov` static method and the distribution to be tested.

# %%
result = ot.FittingTest.Kolmogorov(sample, distribution, 0.01)
print("Conclusion=", result.getBinaryQualityMeasure(), "P-value=", result.getPValue())

# %%
# Test succeeded ?

# %%
result.getBinaryQualityMeasure()

# %%
# P-Value associated to the risk

# %%
result.getPValue()

# %%
# Threshold associated to the test.

# %%
result.getThreshold()

# %%
# Observed value of the statistic.

# %%
result.getStatistic()

# %%
# Case 2 : the distribution parameters are estimated from the sample.
# -------------------------------------------------------------------
#
# In the case where the parameters of the distribution are estimated from the sample,
# we must use the `Lilliefors` static method and the distribution factory to be tested.

# %%
ot.ResourceMap.SetAsUnsignedInteger("FittingTest-LillieforsMaximumSamplingSize", 1000)

# %%
distributionFactory = ot.NormalFactory()

# %%
dist, result = ot.FittingTest.Lilliefors(sample, distributionFactory, 0.01)
print("Conclusion=", result.getBinaryQualityMeasure(), "P-value=", result.getPValue())

# %%
dist

# %%
# Test succeeded ?

# %%
result.getBinaryQualityMeasure()

# %%
# P-Value associated to the risk

# %%
result.getPValue()

# %%
# Threshold associated to the test.

# %%
result.getThreshold()

# %%
# Observed value of the statistic.

# %%
result.getStatistic()



#!/usr/bin/env python

import openturns as ot
import pandas as pd

# check indexation compatibility
columns = ot.Description(["x1", "x2", "x3"])
df1 = pd.DataFrame(index=["R1"], columns=columns)
columns = ot.Point([1.0, 2.0, 3.0])
df2 = pd.DataFrame(index=["R2"], columns=columns)
columns = ot.Indices([1, 2, 3])
df3 = pd.DataFrame(index=["R3"], columns=columns)

sample = ot.Normal(3).getSample(10)

# Sample => Dataframe
df = sample.asDataFrame()
print(df)
assert df.shape == (sample.getSize(), sample.getDimension()), "wrong df shape"

# Dataframe => Sample
sample2 = ot.Sample.BuildFromDataFrame(df)
print(sample2)
assert (sample2.getSize(), sample2.getDimension()) == df.shape, "wrong sample shape"



import openturns as ot
import openturns.viewer as otv

f = ot.SymbolicFunction(["x"], ["x * sin(x)"])
sampleX = [[1.0], [2.0], [3.0], [4.0], [5.0], [6.0], [7.0], [8.0]]
sampleY = f(sampleX)
basis = ot.Basis([ot.SymbolicFunction(["x"], ["1"])])
covarianceModel = ot.SquaredExponential([1.0])
covarianceModel.setActiveParameter([])
fitter_algo = ot.GaussianProcessFitter(sampleX, sampleY, covarianceModel, basis)
fitter_algo.run()
algo = ot.GaussianProcessRegression(fitter_algo.getResult())
algo.run()
result = algo.getResult()
fhat = result.getMetaModel()

graph = f.draw(0.0, 10.0)
graph.add(fhat.draw(0.0, 10.0))
graph.add(ot.Cloud(sampleX, sampleY))

# Conditional variance
gccc = ot.GaussianProcessConditionalCovariance(result)
# Vriance per marginal & extract variance component
newSampleX = ot.Sample.BuildFromPoint(
    [0.0, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5]
)
newSampleX.add(sampleX)
newSampleX = newSampleX.sort()
newSampleY = fhat(newSampleX)
conditionalVariance = gccc.getConditionalMarginalVariance(newSampleX) + [1e-15]
conditionalSigma = ot.SymbolicFunction("x", "sqrt(x)")(conditionalVariance)
alpha = 0.1
bilateralCI = ot.Normal().computeBilateralConfidenceInterval(1.0 - alpha)
lowerCI = newSampleY + conditionalSigma * bilateralCI.getLowerBound()
upperCI = newSampleY + conditionalSigma * bilateralCI.getUpperBound()
boundsPoly = ot.Polygon.FillBetween(newSampleX, lowerCI, upperCI)
boundsPoly.setColor("green")
boundsPoly_legend = "%d%% C.I." % ((1.0 - alpha) * 100)
graph.add(boundsPoly)
graph.setLegends(["model", "meta model", "sample", boundsPoly_legend])
graph.setLegendPosition("upper left")
graph.setTitle("y(x)=x * sin(x)")
otv.View(graph, figure_kw={"figsize": (8, 4)})



"""
Polynomial chaos exploitation
=============================

The goal of this example is to create a polynomial chaos expansion using the
:class:`~openturns.FunctionalChaosAlgorithm` class and see the different methods
of the :class:`~openturns.FunctionalChaosResult` class.
In order to understand the different results, let us review some notations associated to the
polynomial chaos expansion.

We first introduce the physical input, output and model:

- the random vector :math:`\\vect{X}` is the physical input random vector,
- the output random vector :math:`\\vect{Y}` is the output of the physical model,
- the physical model :math:`g` is a function of the physical input:

.. math:: \\vect{Y} = g(\\vect{X}).

Then we introduce the iso-probabilistic transformation:

- the random vector :math:`\\vect{Z}` is the standardized input random vector,
- the transformation :math:`T` is the iso-probabilistic transformation mapping
  from the physical input to the standardized input:

.. math:: \\vect{Z} = T(\\vect{X}),

- the standard distribution :math:`\\mu` of the standardized random vector :math:`\\vect{Z}`.

We expand the model on the multivariate basis:

- the full (i.e. unselected) multivariate basis is :math:`\\left(\\Psi_k\\right)_{k \\in \\mathbb{N}}`,
- the composition of each polynomial of the truncated multivariate basis :math:`\\Psi_k`,
- the full set of coefficients of the polynomial expansion is the set :math:`(\\vect{a}_k)_{k \\in \\mathbb{N}}`,
- the composed model :math:`h` is a function of the standardized random vector :math:`\\vect{Z}`:

.. math:: \\vect{Y} = h(\\vect{Z}) = g \\circ T^{-1}(\\vect{Z})

Then we use a model selection method (e.g. from the :class:`~openturns.LARS` class):

- the set :math:`\\mathcal{K} \\subseteq \\mathbb{N}` is the set of indices corresponding to the
  result of the selection method,
- the truncated (i.e. selected) multivariate basis is :math:`\\left(\\Psi_k\\right)_{k \\in \\mathcal{K}}`,
- the selected set of coefficients of the polynomial expansion is the set :math:`(\\vect{a}_k)_{k \\in \\mathcal{K}}`,
- the composed meta model :math:`\\hat{h}` is the function
  of the standardized variables based on the truncated
  multivariate basis :math:`\\left(\\Psi_k\\right)_{k \\in \\mathcal{K}}`.
- the meta model is a function of the physical input:

.. math:: \\vect{Y} = \\hat{g} (\\vect{X}) = \\hat{h} \\circ T(\\vect{X}).


Based on the previous definitions, the composed model is:

.. math::

    h(\\vect{Z}) =  \\sum_{k \\in \\mathbb{N}} \\vect{a}_k \\Psi_k(\\vect{Z}),

the composed metamodel is:

.. math::

    \\hat{h}(\\vect{Z}) = \\sum_{k \\in \\mathcal{K}} \\vect{a}_k \\Psi_k(\\vect{Z}),

and the metamodel is:

.. math::

    \\hat{g}(\\vect{X}) = \\sum_{k \\in \\mathcal{K}} \\vect{a}_k \\Psi_k \\circ T(\\vect{X}).

The three previous mathematical functions are implemented as instances of the
:class:`~openturns.Function` class.
"""

# %%
# Create the polynomial chaos expansion
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# %%
import openturns as ot


# %%
# Prepare some input and output samples.
# We define a model which has two inputs and two outputs.
# Then we define a Normal input random vector with independent marginals,
# and we generate a sample from the input random vector.
# Finally, we evaluate the output sample from the model.
dimension = 2
input_names = ["x1", "x2"]
formulas = ["cos(x1 + x2)", "(x2 + 1) * exp(x1 - 2 * x2)"]
model = ot.SymbolicFunction(input_names, formulas)
distribution = ot.Normal(dimension)
x = distribution.getSample(30)
y = model(x)

# %%
# Create a functional chaos algorithm.
algo = ot.FunctionalChaosAlgorithm(x, y)

# %%
# The previous constructor is the simplest since it has only two input arguments.
# In this case, the algorithm has to retrieve the distribution from the `x`
# sample, which can be difficult, especially if the sample size is small.
# Notice, however, that the input distribution is known in our simple case.
# This is why we update the previous script and give the input distribution as a third
# input argument of the constructor.
algo = ot.FunctionalChaosAlgorithm(x, y, distribution)
algo.run()

# %%
# Get the result.
result = algo.getResult()
result

# %%
# Get the coefficients
# ~~~~~~~~~~~~~~~~~~~~

# %%
# In the next cells, we observe several methods of the `result` object.
# First, we get the polynomial chaos coefficients.
result.getCoefficients()

# %%
# The coefficients of the i-th output marginal.
i = 1
result.getCoefficients()[i]

# %%
# Get the indices of the selected polynomials.
subsetK = result.getIndices()
subsetK

# %%
# Get the composition of the polynomials
# of the truncated multivariate basis.
for i in range(subsetK.getSize()):
    print(
        "Polynomial number ",
        i,
        " in truncated basis <-> polynomial number ",
        subsetK[i],
        " = ",
        ot.LinearEnumerateFunction(dimension)(subsetK[i]),
        " in complete basis",
    )

# %%
# Get the basis
# ~~~~~~~~~~~~~

# %%
# Get the multivariate basis
# as a collection of :class:`~openturns.Function` objects.
reduced = result.getReducedBasis()

# %%
# Get the orthogonal basis.
orthgBasis = result.getOrthogonalBasis()

# %%
# Get the standard distribution :math:`\mu` of the standardized
# random vector :math:`\vect{Z}`.
orthgBasis.getMeasure()

# %%
# Get the metamodel
# ~~~~~~~~~~~~~~~~~

# %%
# Get the composed meta model :math:`\hat{h}` which is
# the model of the standardized random vector :math:`\vect{Z}`
# within the reduced polynomials basis.
result.getComposedMetaModel()

# %%
# Get the meta model :math:`\hat{g}` which is the composed meta model composed with the
# iso probabilistic transformation :math:`T`.
result.getMetaModel()

# %%
# Get the projection method
# ~~~~~~~~~~~~~~~~~~~~~~~~~

# %%
# Get the projection strategy: this is the method to estimate
# the coefficients, i.e. regression or integration.
algo.getProjectionStrategy()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# 1d
mesh = ot.IntervalMesher([10]).build(ot.Interval(-2.0, 2.0))
graph = mesh.draw()

# 2d
mesh = ot.IntervalMesher([10] * 2).build(ot.Interval([-2.0] * 2, [2.0] * 2))
graph = mesh.draw()

# 3d
mesh = ot.IntervalMesher([10] * 3).build(ot.Interval([-2.0] * 3, [2.0] * 3))
graph = mesh.draw3D()

for drawEdge in [True, False]:
    for shading in [True, False]:
        for rho in [1.0, 0.8]:
            for alpha in [1.0, 0.5]:
                faceColor = ot.Drawable.ConvertToRGBA(
                    ot.Drawable.ConvertFromName(
                        ot.ResourceMap.GetAsString("Mesh-FaceColor")
                    )
                )
                faceColor[3] = int(255 * alpha)
                ot.ResourceMap.SetAsString(
                    "Mesh-FaceColor",
                    ot.Drawable.ConvertFromRGBA(*faceColor),
                )
                edgeColor = ot.Drawable.ConvertToRGBA(
                    ot.Drawable.ConvertFromName(
                        ot.ResourceMap.GetAsString("Mesh-EdgeColor")
                    )
                )
                edgeColor[3] = int(255 * alpha)
                ot.ResourceMap.SetAsString(
                    "Mesh-EdgeColor",
                    ot.Drawable.ConvertFromRGBA(*edgeColor),
                )

                graph = mesh.draw3D(drawEdge, 0.2, 0.5, 0.9, shading, rho)
                graph.setTitle(f"{drawEdge=}, {shading=}, {rho=}, {alpha=}")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Hann filter
myHanningFilter = ot.Hann()
print("myHanningFilter = ", myHanningFilter)

# Hamming filter
myHammingFilter = ot.Hamming()
print("myHammingFilter = ", myHammingFilter)

# Evaluation of values between t=0 and t=1 using a step = 0.01
steps = 100
tMin = 0.0
tMax = 1.0
tStep = (tMax - tMin) / steps

for i in range(steps + 1):
    t = tMin + i * tStep
    print("t = %g Hann = %g Hamming = %g" % (t, myHanningFilter(t), myHammingFilter(t)))



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

mean = [3.0, 2.0, 1.0]
sigma = [2.0, 3.0, 4.0]

refCovariances = list()
refCovariances.append(
    ot.CovarianceMatrix(3, [4.0, 0.0, 0.0, 0.0, 9.0, 0.0, 0.0, 0.0, 16.0])
)
refCovariances.append(
    ot.CovarianceMatrix(3, [4.0, 1.5, 0.0, 1.5, 9.0, 3.0, 0.0, 3.0, 16.0])
)
refCovariances.append(
    ot.CovarianceMatrix(3, [4.0, 1.125, 0.0, 1.125, 9.0, 2.25, 0.0, 2.25, 16.0])
)
refCovariances.append(
    ot.CovarianceMatrix(
        3,
        [
            4.0,
            2.0696999,
            -4.403889,
            2.0696999,
            9.0,
            4.1393998,
            -4.403889,
            4.1393998,
            16.0,
        ],
    )
)
refCovariances.append(
    ot.CovarianceMatrix(
        3, [0.39606657, 0.0, 0.0, 0.0, 0.891149785, 0.0, 0.0, 0.0, 1.584266284]
    )
)

refStandardDeviation = [
    [2, 3, 4],
    [2, 3, 4],
    [2, 3, 4],
    [1.49595080640498, 2.00948748222124, 2.99190161280996],
    [0.62933820074628, 0.94400730111948, 1.25867640149264],
]
refSkewness = [
    [0, 0, 0],
    [0, 0, 0],
    [0, 0, 0],
    [-0.213157049688829, 0.0, 0.213157049689032],
    [0.22711106425, 0.22711106425, 0.22711106425],
]
refKurtosis = [
    [3, 3, 3],
    [3, 3, 3],
    [3, 3, 3],
    [3.11664895604121, 3.03472746922749, 3.11664895604127],
    [2.439305739629, 2.439305739629, 2.439305739629],
]

# Create a collection of distribution attente TUI
aCollection = []
try:
    aCollection[10] = ot.Normal()
except Exception:
    pass

dim = len(mean)
# Create a collection of distribution
aCollection = list()
marginal = ot.Normal(mean[0], sigma[0])
marginal.setName("First")
marginal.setDescription(["One"])
aCollection.append(marginal)
marginal = ot.Normal(mean[1], sigma[1])
marginal.setName("Second")
marginal.setDescription(["Two"])
aCollection.append(marginal)
marginal = ot.Normal(mean[2], sigma[2])
marginal.setName("Third")
marginal.setDescription(["Three"])
aCollection.append(marginal)

# Create a copula
aCopula = ot.IndependentCopula(dim)
aCopula.setName("Independent copula")
cores = [aCopula]
# With a Normal copula
correlation = ot.CorrelationMatrix(dim)
for i in range(1, dim):
    correlation[i - 1, i] = 0.25
anotherCopula = ot.NormalCopula(correlation)
anotherCopula.setName("Normal copula")
cores.append(anotherCopula)
# With a copula which is not a copula by type
atoms = [aCopula, anotherCopula]
cores.append(ot.Mixture(atoms, [0.25, 0.75]))
# With a non-copula core
cores.append(ot.UniformOrderStatistics(dim))
# With a core which support is strictly included in the unit cube
cores.append(ot.KernelMixture(ot.Beta(2.0, 3.0, 0.2, 0.8), [1.0] * dim, [[0.0] * dim]))
ot.ResourceMap.SetAsBool("JointDistribution-UseGenericCovarianceAlgorithm", True)
for nCore in range(len(cores)):
    print("\n\n")
    # Instantiate one distribution object
    distribution = ot.JointDistribution(aCollection, cores[nCore])
    distribution.setName("myDist")
    print("Distribution", repr(distribution))
    print("Distribution")
    print(distribution)
    print("Distribution (Markdown)")
    print(distribution.__repr_markdown__())
    print("Parameters", distribution.getParametersCollection())
    print("nCore=", nCore)

    # too slow for Mixture/KernelMixture
    if "Mixture" not in distribution.getCore().getImplementation().getName():
        print("entropy=%.5e" % distribution.computeEntropy())
        print(
            "entropy (MC)=%.5e"
            % -distribution.computeLogPDF(
                distribution.getSample(1000000)
            ).computeMean()[0]
        )
    print("Mean ", distribution.getMean())
    precision = ot.PlatformInfo.GetNumericalPrecision()
    print("Covariance ")
    if nCore != 3:
        covariance = distribution.getCovariance()
        covariance.checkSymmetry()
        ott.assert_almost_equal(covariance, refCovariances[nCore])
    # Is this distribution an elliptical distribution?
    print("Elliptical distribution= ", distribution.isElliptical())

    # Has this distribution an elliptical copula?
    print("Elliptical copula= ", distribution.hasEllipticalCopula())

    # Has this distribution an independent copula?
    print("Independent copula= ", distribution.hasIndependentCopula())

    # Test for realization of distribution
    oneRealization = distribution.getRealization()
    print("oneRealization=", oneRealization)
    # Test for sampling
    size = 10
    oneSample = distribution.getSample(size)
    print("oneSample=", oneSample)

    # Test for sampling
    size = 10000
    anotherSample = distribution.getSample(size)
    print("anotherSample mean=", anotherSample.computeMean())
    print("anotherSample covariance=", anotherSample.computeCovariance())

    # Define a point
    zero = [0.0] * dim

    # Show PDF and CDF of zero point
    zeroPDF = distribution.computePDF(zero)
    zeroCDF = distribution.computeCDF(zero)
    print("Zero point= ", zero, " pdf=%.5e" % zeroPDF, " cdf=%.5e" % zeroCDF)
    # Get 95% quantile
    quantile = distribution.computeQuantile(0.95)
    print("Quantile=", quantile)
    print("CDF(quantile)=%.5e" % distribution.computeCDF(quantile))

    # Extract the marginals
    for i in range(dim):
        margin = distribution.getMarginal(i)
        print("margin=", margin)
        print("margin PDF=%.5e" % margin.computePDF([0.0]))
        print("margin CDF=%.5e" % margin.computeCDF([0.0]))
        print("margin quantile=", margin.computeQuantile(0.95))
        print("margin realization=", margin.getRealization())

    # Extract a 2-D marginal
    indices = [1, 0]
    print("indices=", indices)
    margins = distribution.getMarginal(indices)
    print("margins=", margins)
    print("margins PDF=%.5e" % margins.computePDF([0.0] * 2))
    print("margins CDF=%.5e" % margins.computeCDF([0.0] * 2))
    quantile = margins.computeQuantile(0.5)
    print("margins quantile=", quantile)
    print("margins CDF(quantile)=%.5e" % margins.computeCDF(quantile))
    print("margins realization=", margins.getRealization())
    x = 0.6
    y = [0.2] * (dim - 1)
    print("conditional PDF=%.5e" % distribution.computeConditionalPDF(x, y))
    print("conditional CDF=%.5e" % distribution.computeConditionalCDF(x, y))
    print("conditional quantile=%.5e" % distribution.computeConditionalQuantile(x, y))
    pt = [i + 1.5 for i in range(dim)]
    print(
        "sequential conditional PDF=", distribution.computeSequentialConditionalPDF(pt)
    )
    resCDF = distribution.computeSequentialConditionalCDF(pt)
    print("sequential conditional CDF(", pt, ")=", resCDF)
    print(
        "sequential conditional quantile(",
        resCDF,
        ")=",
        distribution.computeSequentialConditionalQuantile(resCDF),
    )

    # Confidence regions
    if distribution.getDimension() <= 2:
        (
            levelSet,
            threshold,
        ) = distribution.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
        print("Minimum volume interval=", levelSet)
        print("threshold=%.5e" % threshold)
        levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
        print("Minimum volume level set=", levelSet)
        print("beta=%.5e" % beta)
        (
            interval,
            beta,
        ) = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
        print("Bilateral confidence interval=", interval)
        print("beta=%.5e" % beta)
        (
            interval,
            beta,
        ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
            0.95, False
        )
        print("Unilateral confidence interval (lower tail)=", interval)
        print("beta=%.5e" % beta)
        (
            interval,
            beta,
        ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
            0.95, True
        )
        print("Unilateral confidence interval (upper tail)=", interval)
        print("beta=%.5e" % beta)
    # Moments other than mean and covariance
    standardDeviation = distribution.getStandardDeviation()
    ott.assert_almost_equal(standardDeviation, refStandardDeviation[nCore])
    skewness = distribution.getSkewness()
    ott.assert_almost_equal(skewness, refSkewness[nCore])
    kurtosis = distribution.getKurtosis()
    ott.assert_almost_equal(kurtosis, refKurtosis[nCore])
    anotherSample = distribution.getSample(size)
    print("anotherSample mean=", anotherSample.computeMean())
    print("anotherSample covariance=", anotherSample.computeCovariance())

# Create and print a complex distribution
aCollection2 = list()
marginalN = ot.Normal(0.0, 1.0)
marginalN.setName("First")
marginalN.setDescription(["One"])
aCollection2.append(marginalN)
marginalU = ot.Uniform(12345.6, 123456.7)
marginalU.setName("Second")
marginalU.setDescription(["Two"])
aCollection2.append(marginalU)
marginalTN = ot.TruncatedDistribution(ot.Normal(2.0, 1.5), 1.0, 4.0)
marginalTN.setName("Third")
marginalTN.setDescription(["Three"])
aCollection2.append(marginalTN)
distribution2 = ot.JointDistribution(aCollection2)
distribution2.setName("myDist2")
print("Distribution ")
print(distribution2)
print("Distribution (Markdown)")
print(distribution2.__repr_markdown__())

# test transfo
sample = distribution.getSample(10)
print(sample)
sample_iso = distribution.getIsoProbabilisticTransformation()(sample)
print(sample_iso)
sample_inv = distribution.getInverseIsoProbabilisticTransformation()(sample_iso)
print(sample_inv)
x = 0.6
y = [0.2] * (dim - 1)
print("conditional PDF=%.6f" % distribution.computeConditionalPDF(x, y))
print("conditional CDF=%.6f" % distribution.computeConditionalCDF(x, y))
print("conditional quantile=%.6f" % distribution.computeConditionalQuantile(x, y))
pt = [i + 1.5 for i in range(dim)]
print("sequential conditional PDF=", distribution.computeSequentialConditionalPDF(pt))
resCDF = distribution.computeSequentialConditionalCDF(pt)
print("sequential conditional CDF(", pt, ")=", resCDF)
print(
    "sequential conditional quantile(",
    resCDF,
    ")=",
    distribution.computeSequentialConditionalQuantile(resCDF),
)

# comparison
d = ot.JointDistribution([ot.Uniform()] * 2, ot.ClaytonCopula(2.5))
d1 = ot.MaximumDistribution(d)
d2 = ot.MaximumDistribution(d)
assert d1.getDistribution() == d2.getDistribution(), "comp1"
assert d1 == d2, "comp2"

# check for lost description
dist_a = ot.Normal()
dist_a.setDescription(["a"])
dist_b = ot.Normal()
dist_b.setDescription(["b"])
dist_list = [dist_a, dist_b] + [ot.Normal()] * 3
composed = ot.JointDistribution(dist_list)
assert composed.getDescription() == ["a", "b", "X0", "X1", "X2"], "wrong description"

# Create and print a composed distribution with different
# complexities and print them
aCollection = [
    ot.Uniform(),
    ot.TruncatedDistribution(ot.Normal(2.0, 1.5), 1.0, 4.0),
    ot.Normal(),
]
distribution = ot.JointDistribution(aCollection)
print("Distribution = ")
print(distribution)
print("Distribution (Markdown) = ")
print(distribution.__repr_markdown__())
print("Distribution (HTML) = ")
print(distribution._repr_html_())

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.run()

# Check if one can detect if the distribution is elliptical
# Normal marginals, independent copula
distribution = ot.JointDistribution([ot.Normal(1.0, 2.0), ot.Normal(-2.0, 1.0)])
print(
    "Normal marginals, independent copula, isElliptical?", distribution.isElliptical()
)
# Normal marginals, normal copula
R = ot.CorrelationMatrix(2, [1.0, 0.5, 0.5, 1.0])
distribution = ot.JointDistribution(
    [ot.Normal(1.0, 2.0), ot.Normal(-2.0, 1.0)], ot.NormalCopula(R)
)
print("Normal marginals, normal copula, isElliptical?", distribution.isElliptical())
# Normal marginals, non-normal copula
distribution = ot.JointDistribution(
    [ot.Normal(1.0, 2.0), ot.Normal(-2.0, 1.0)], ot.ClaytonCopula(1.0)
)
print("Normal marginals, non-normal copula, isElliptical?", distribution.isElliptical())
# Student marginals, Student copula, all with the same nu
nu = 3.5
distribution = ot.JointDistribution(
    [ot.Student(nu, 1.0, 2.0), ot.Student(nu, -2.0, 1.0)], ot.StudentCopula(nu, R)
)
print(
    "Student marginals, Student copula, same nu, isElliptical?",
    distribution.isElliptical(),
)
distribution = ot.JointDistribution(
    [ot.Student(nu, 1.0, 2.0), ot.Student(nu, -2.0, 1.0)], ot.StudentCopula(nu + 1.0, R)
)
print(
    "Student marginals same nu, Student copula different nu, isElliptical?",
    distribution.isElliptical(),
)
distribution = ot.JointDistribution(
    [ot.Student(nu + 1.0, 1.0, 2.0), ot.Student(nu, -2.0, 1.0)],
    ot.StudentCopula(nu + 1.0, R),
)
print(
    "Student marginals different nu, Student copula same nu as first marginal, isElliptical?",
    distribution.isElliptical(),
)



"""
Draw a field
============
"""

# %%
# In this example, we show the different ways of drawing a :class:`~openturns.Field`
# that is a realization of a process.
#
# We consider the following fields:
#
# - **Case 1**: A field defined on a  2-d mesh with 2-d values,
# - **Case 2**: A field defined on a  2-d mesh with scalar values,
# - **Case 3**: A field defined on a  1-d mesh with  2-d values,
# - **Case 4**: A field defined on a  1-d mesh with  scalar values.
#

# %%
import openturns as ot
import openturns.viewer as otv

# sphinx_gallery_thumbnail_number = 2

# %%
# Case 1: A field defined on a  2-d mesh with 2-d values
# ------------------------------------------------------
# We consider a normal process defined on :math:`\cD = [0,1]^2`, with 2-d output values, defined by a
# covariance function which is a :class:`~openturns.TensorizedCovarianceModel` built as the tensorization of
# two  :class:`~openturns.MaternModel` covariance functions.
# The domain :math:`\cD` is discretized by a regular grid, using the class
# :class:`~openturns.IntervalMesher`.
cov_model = ot.TensorizedCovarianceModel([ot.MaternModel([0.1] * 2, [0.01], 2.5)] * 2)
mesh = ot.IntervalMesher([30] * 2).build(ot.Interval([0.0] * 2, [1.0] * 2))
normal_proc = ot.GaussianProcess(cov_model, mesh)
field = normal_proc.getRealization()

# %%
# We first draw the field with arrows fixed on each vertex of the mesh.
g_1 = field.draw()
g_1.setTitle(r"A field on 2-d mesh and 2-d values")
g_1.setXTitle(r"$x_0$")
g_1.setYTitle(r"$x_1$")

# %%
# In order to change the arrows, we can modify the :class:`~openturns.ResourceMap` keys associated to the
# class :class:`~openturns.Field`.
ot.ResourceMap.SetAsScalar("Field-ArrowRatio", 0.05)
ot.ResourceMap.SetAsScalar("Field-ArrowScaling", 0.8)
g_2 = field.draw()

g_2.setTitle("The same field with other arrows parameters")
g_2.setXTitle(r"$x_0$")
g_2.setYTitle(r"$x_1$")

# %%
# We can compare both figures.
grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, g_1)
grid.setGraph(0, 1, g_2)
view = otv.View(grid)

# %%
# We can also use the output values to deform the initial mesh and draw the deformed mesh.
# First, we recover the default values of the ResourceMap keys.
ot.ResourceMap.SetAsScalar("Field-ArrowRatio", 0.01)
ot.ResourceMap.SetAsScalar("Field-ArrowScaling", 1)
deformed_field = field.asDeformedMesh()
g_def = deformed_field.draw()
g_def.setTitle("The same field as deformed mesh")
g_def.setXTitle(r"$x_0$")
g_def.setYTitle(r"$x_1$")
g_def.setLegendPosition("")

# %%
# We can compare both figures.
grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, g_1)
grid.setGraph(0, 1, g_def)
view = otv.View(grid)

# %%
# Case 2: A field defined on a  2-d mesh with scalar values
# ---------------------------------------------------------
# We consider a normal process defined on :math:`\cD = [0,1]^2`, with scalar output values, defined by a
# covariance function which is a :class:`~openturns.MaternModel`.
# The domain :math:`\cD` is discretized by a regular grid, using the class :class:`~openturns.IntervalMesher`.
cov_model = ot.MaternModel([0.15] * 2, [0.01], 2.5)
mesh = ot.IntervalMesher([35] * 2).build(ot.Interval([0.0] * 2, [1.0] * 2))
normal_proc = ot.GaussianProcess(cov_model, mesh)
field = normal_proc.getRealization()

# %%
# We draw the field by filling each cell of the mesh with a color.
g_1 = field.draw()
g_1.setTitle("A field on 2-d mesh  and scalar values")
g_1.setXTitle(r"$x_0$")
g_1.setYTitle(r"$x_1$")


# %%
# We can also draw iso-value lines of the output values using interpolation.
g_2 = field.drawMarginal(0, True)
g_2.setTitle("A field on 2-d mesh and scalar values")
g_2.setXTitle(r"$x_0$")
g_2.setYTitle(r"$x_1$")
g_2.setLegendPosition("")

# %%
# We can compare both figures.
grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, g_1)
grid.setGraph(0, 1, g_2)
view = otv.View(grid)

# %%
# Case 3: A field defined on a  1-d mesh with  2-d values
# -------------------------------------------------------
# We consider a :class:`~openturns.RandomWalk` process which distribution is the normal
# distribution of dimension 2 with zero mean and identity covariance matrix.
# The process is defined on :math:`\cD = [0,1]` which is discretized by a regular grid, using
# the class :class:`~openturns.RegularGrid`.
dist_rw = ot.Normal(2)
origin = dist_rw.getMean()
N = 10000
step = 1.0 / N
mesh = ot.RegularGrid(0, step, N)
rw_process = ot.RandomWalk(origin, dist_rw, mesh)
field = rw_process.getRealization()

# %%
# We draw the field by interpolating the output vectors.
g = field.draw()
g.setTitle("A 2-d random walk field")
g.setXTitle(r"$x_0$")
g.setYTitle(r"$x_1$")
view = otv.View(g)

# %%
# Case 4: A field defined on a  1-d mesh with  scalar values
# ----------------------------------------------------------
# We consider a normal process defined on :math:`\cD = [0,1]`, with scalar output values, defined by a
# covariance function which is a :class:`~openturns.MaternModel`.
# The domain :math:`\cD` is discretized by a regular grid, using  :class:`~openturns.IntervalMesher`.
# This process is stationary.
cov_model = ot.MaternModel([0.05], [0.01], 2.5)
mesh = ot.IntervalMesher([1000]).build(ot.Interval(0.0, 1.0))
normal_proc = ot.GaussianProcess(cov_model, mesh)
field = normal_proc.getRealization()

# %%
# We draw its trajectory without interpolating the output values.
g_1 = field.draw()
g_1.setTitle("A field on 1-d mesh and scalar values - Stationary process")
g_1.setXTitle(r"$x_0$")
g_1.setYTitle(r"$x_1$")

# %%
# We can also interpolate the output values.
g_2 = field.drawMarginal(0, True)
g_2.setTitle("The same field with interpolation")
g_2.setXTitle(r"$x_0$")
g_2.setYTitle(r"$x_1$")
g_2.setLegendPosition("")

# %%
# We can compare both figures.
grid = ot.GridLayout(1, 2)
grid.setGraph(0, 0, g_1)
grid.setGraph(0, 1, g_2)
view = otv.View(grid)

# %%
# We consider now a normal process  which is not stationary. We use the
# :class:`~openturns.FractionalBrownianMotionModel`:
#
# - when the Hurst exponent is close to 1, the trajectory is smooth,
# - when the Hurst exponent is equal to 1/2, the trajectory is Brownian motion (Wiener process),
# - when the Hurst exponent is close to 0, the process tends to a White noise process.
#
scale = 0.01
amplitude = 1.5
hurst_exponent = 0.75
cov_model = ot.FractionalBrownianMotionModel(scale, amplitude, hurst_exponent)
mesh = ot.IntervalMesher([1000]).build(ot.Interval(0.0, 1.0))
normal_proc = ot.GaussianProcess(cov_model, mesh)
field = normal_proc.getRealization()
g_3 = field.drawMarginal(0, True)
g_3.setTitle("A field on 1-d mesh and scalar values - Non stationary process")
g_3.setXTitle(r"$x_0$")
g_3.setYTitle(r"$x_1$")

# %%
# We can compare both processes.
grid = ot.GridLayout(1, 2)
g_2.setTitle("A field on 1-d mesh and scalar values - Stationary process")
grid.setGraph(0, 0, g_2)
grid.setGraph(0, 1, g_3)
view = otv.View(grid)

# %%
# We draw here a trajectory from a FractionalBrownianMotionModel with different Hurst exponent.
grid = ot.GridLayout(1, 3)
hurst_exponent_list = [0.99, 0.50, 0.001]
for index, hurst_exponent in enumerate(hurst_exponent_list):
    cov_model = ot.FractionalBrownianMotionModel(scale, amplitude, hurst_exponent)
    normal_proc = ot.GaussianProcess(cov_model, mesh)
    field = normal_proc.getRealization()
    g = field.drawMarginal(0, True)
    g.setTitle(
        r"Fractional Brownian Motion Model, Hurst exponent = " + str(hurst_exponent)
    )
    g.setXTitle(r"$x_0$")
    g.setYTitle(r"$x_1$")
    grid.setGraph(0, index, g)
view = otv.View(grid)

# %%
# Display all figures
otv.View.ShowAll()



"""
Test the copula
===============
"""

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Copula fitting test using Kendall plot
# --------------------------------------
#
# We first perform a visual goodness-of-fit test for a copula with the Kendall plot test.

# %%
# We create two samples of size 500.
N = 500

# %%
dist1 = ot.JointDistribution([ot.Normal()] * 2, ot.GumbelCopula(3.0))
sample1 = dist1.getSample(N)
sample1.setName("sample1")

# %%
dist2 = ot.JointDistribution([ot.Normal()] * 2, ot.ClaytonCopula(0.2))
sample2 = dist2.getSample(N)
sample2.setName("sample2")

# %%
# We change the parameter for the evaluation of :math:`\Expect{W_i}` thanks to the :class:`~openturns.ResourceMap` :
ot.ResourceMap.SetAsUnsignedInteger("VisualTest-KendallPlot-MonteCarloSize", 25)

# %%
# We can test a specific copula model for a given sample,
copula_test = ot.GumbelCopula(3)
graph = ot.VisualTest.DrawKendallPlot(sample1, copula_test)
view = otv.View(graph)

# %%
# or test whether two samples have the same copula model
graph = ot.VisualTest.DrawKendallPlot(sample1, sample2)
view = otv.View(graph)

# %%
# The first test gives a positive result as the blue curve is near the identity line which is not the case for the second test.


# %%
# Graphical copula validation
# ---------------------------
#
# In this paragraph we visualize an estimated copula versus the data in the rank space.
#

# %%
# First we create data
marginals = [ot.Normal()] * 2
dist = ot.JointDistribution(marginals, ot.ClaytonCopula(3))
N = 500
sample = dist.getSample(N)

# %%
# We build a estimate copula from the previous sample using the :class:`~openturns.ClaytonCopulaFactory` :
estimated = ot.ClaytonCopulaFactory().build(sample)

# %%
# We represent data as a cloud in the rank space :
ranksTransf = ot.MarginalTransformationEvaluation(
    marginals, ot.MarginalTransformationEvaluation.FROM
)
rankSample = ranksTransf(sample)
rankCloud = ot.Cloud(rankSample, "blue", "plus", "sample")

# %%
# We can plot the graph with rank sample and estimated copula :
myGraph = ot.Graph("Parametric estimation of the copula", "X", "Y")
myGraph.setLegendPosition("upper left")
myGraph.setLegendPosition("lower right")
myGraph.add(rankCloud)

# %%
# and draw the iso-curves of the estimated copula
minPoint = [0.0] * 2
maxPoint = [1.0] * 2
pointNumber = [201] * 2
graphCop = estimated.drawPDF(minPoint, maxPoint, pointNumber)
contour_estCop = graphCop.getDrawable(0)
# Erase the labels of the iso-curves
contour_estCop.setDrawLabels(False)
# Change the levels of the iso-curves
nlev = 21
levels = [0.0] * nlev
for i in range(nlev):
    levels[i] = 0.25 * nlev / (nlev - i)
contour_estCop.setLevels(levels)
# Change the legend of the curves
contour_estCop.setLegend("Gumbel copula")
# Change the color of the iso-curves
contour_estCop.setColor("red")
# Add the iso-curves graph into the cloud one
myGraph.add(contour_estCop)
view = otv.View(myGraph)


# %%
# Display figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# DEFAULT CONSTRUCTOR AND STRING CONVERTER
print("test : default constructor and string converter")

# Default constructor
hermitianMatrix0 = ot.HermitianMatrix()

# String converter
print("hermitianMatrix0 = ", repr(hermitianMatrix0))

# CONSTRUCTOR WITH SIZE, OPERATOR() AND STRING CONVERTER
print("test : constructor with size, operator() and string converter")

# Constructor with size
hermitianMatrix1 = ot.HermitianMatrix(2)

# Check operator() methods
hermitianMatrix1[0, 0] = 1.0
hermitianMatrix1[1, 0] = 3.0 + 1.0j
hermitianMatrix1[1, 1] = 4.0

# String converter
print("hermitianMatrix1 = ", repr(hermitianMatrix1))

# COPY CONSTRUCTOR AND STRING CONVERTER
print("test : copy constructor and string converter")

# Copy constructor
hermitianMatrix2 = ot.HermitianMatrix(hermitianMatrix1)

# String converter
print("hermitianMatrix2 = ", repr(hermitianMatrix2))

# GET DIMENSIONS METHODS
print("test : get dimensions methods")

# Get dimension methods
print("hermitianMatrix1's nbRows = ", hermitianMatrix1.getNbRows())
print("hermitianMatrix1's nbColumns = ", hermitianMatrix1.getNbColumns())

# CONJUGATE METHOD
print("test  : conjugate method")

# Check conjugate method
hermitianMatrix4 = hermitianMatrix1.conjugate()
print("hermitianMatrix1 conjugate = ", repr(hermitianMatrix4))

#  ADDITION METHOD
print("test : addition method")

# Check addition method : we check the operator and the symmetry of the
# operator, thus testing the comparison operator
sum1 = hermitianMatrix1 + hermitianMatrix4
sum2 = hermitianMatrix4 + hermitianMatrix1
print("sum1 = ", repr(sum1))
print("sum2 = ", repr(sum2))
print("sum1 equals sum2 = ", sum1 == sum2)

# SUBTRACTION METHOD
print("test : subtraction method")

# Check subtraction method
diff = hermitianMatrix1 - hermitianMatrix4
print("diff = ", repr(diff))

#  MATRIX MULTIPLICATION METHOD
print("test : matrix multiplication method")

# Check multiplication method
prod = hermitianMatrix1 * hermitianMatrix4
print("prod = ", repr(prod))

# MULTIPLICATION WITH A NUMERICAL POINT METHOD
print("test : multiplication with a numerical point method")

# Create the numerical point
pt = ot.Point()
pt.add(1.0)
pt.add(2.0)
print("pt = ", repr(pt))

# Check the product method
ptResult = hermitianMatrix1 * pt
print("ptResult = ", repr(ptResult))

# MULTIPLICATION AND DIVISION BY A NUMERICAL SCALAR METHODS
print("test : multiplication and division by a numerical scalar methods")

# Check the multiplication method
s = 3.0 + 2j
scalprod = hermitianMatrix1 * s

print("scalprod = ", repr(scalprod))

# Check the division method
scaldiv1 = hermitianMatrix1 / s
scaldiv2 = hermitianMatrix1 / s
print("scaldiv1 = ", repr(scaldiv1))
print("scaldiv2 = ", repr(scaldiv2))
print("scaldiv1 equals scaldiv2 = ", (scaldiv1 == scaldiv2))

# ISEMPTY METHOD
print("test : isEmpty method")

# Check method isEmpty
hermitianMatrix5 = ot.HermitianMatrix()
hermitianMatrix6 = ot.HermitianMatrix()
print("hermitianMatrix0 is empty = ", hermitianMatrix0.isEmpty())
print("hermitianMatrix1 is empty = ", hermitianMatrix1.isEmpty())
print("hermitianMatrix5 is empty = ", hermitianMatrix5.isEmpty())



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Normal(4)
size = 10
experiment = ot.LHSExperiment(distribution, size)
print("experiment = ", experiment)
sample, weights = experiment.generateWithWeights()
print("sample = ", repr(sample))
print("weights = ", repr(weights))
print("sample2 = ", experiment.generate())
experiment = ot.LHSExperiment(distribution, size, True, True)
print("sample  = ", experiment.generate())
print("sample2 = ", experiment.generate())
experiment = ot.LHSExperiment(distribution, size, False, False)
print("sample  = ", experiment.generate())
print("sample2 = ", experiment.generate())
experiment = ot.LHSExperiment(distribution, size, True, False)
print("sample  = ", experiment.generate())
print("sample2 = ", experiment.generate())
experiment = ot.LHSExperiment(size)
print("sample  = ", experiment.generate())
print("sample2 = ", experiment.generate())

# variable size
experiment = ot.LHSExperiment(ot.Normal(4), 10)
for size in [215, 464]:
    experiment.setSize(size)
    X = experiment.generate()

# with correlation
dimension = 3
mu = ot.Point(dimension)
sigma = ot.CovarianceMatrix(dimension)
sigma[0, 2] = -0.5
distribution = ot.Normal(mu, sigma)
sampleSize = 5
experiment = ot.LHSExperiment(distribution, sampleSize)
sample = experiment.generate()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(6)
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-DefaultMaxSize", 4000000
)

# Defining LinearCombinationDistribution
weights = ot.Point(0)
coll = ot.DistributionCollection(0)
coll.add(ot.Gamma(0.5, 1.0))
weights.add(1.0)
coll.add(ot.Gamma(0.5, 1.0))
weights.add(1.0)
coll.add(ot.Gamma(0.5, 1.0))
weights.add(1.0)
coll.add(ot.Gamma(0.5, 1.0))
weights.add(1.0)
coll.add(ot.Gamma(1.0, 1.0))
weights.add(1.0)
distribution = ot.LinearCombinationDistribution(coll, weights)
referenceDistribution = ot.Gamma(3.0, 1.0)
# Compute PDF on regular grid
N = 256
points = ot.Indices(1, N)
mean = distribution.getMean()
sigma = distribution.getStandardDeviation()
xMin = mean - 3.9 * sigma
xMax = mean + 3.9 * sigma
grid = ot.Sample()
print("distribution = ", repr(distribution))
print("range = ", distribution.getRange())
print("mean = ", distribution.getMean())
print("cov = ", distribution.getCovariance())
print("sigma = ", distribution.getStandardDeviation())
print("xMin = ", xMin)
print("xMax = ", xMax)
result, grid = distribution.computePDF(xMin, xMax, points)
print("x;PDF;reference")
for i in range(grid.getSize()):
    ref = referenceDistribution.computePDF(grid[i])
    print("%.6g;%.6g;%.6g" % (grid[i][0], result[i][0], ref))



#! /usr/bin/env python

import openturns as ot

# Defining parameters
dimension = 5
size = 100

# Build OT LHS algorithm
lhs = ot.LHSExperiment(ot.JointDistribution([ot.Uniform()] * dimension), size)

# Generate design without optimization
design = lhs.generate()

# Defining space fillings
spaceFillingC2 = ot.SpaceFillingC2()
spaceFillingMinDist = ot.SpaceFillingMinDist()
spaceFillingPhiP = ot.SpaceFillingPhiP()
spaceFillingPhiP50 = ot.SpaceFillingPhiP(50)

# print the criteria on this design
print(
    "C2=%f MinDist=%f PhiP=%f, PhiP(50)=%f"
    % tuple(
        [
            sf.evaluate(design)
            for sf in [
                spaceFillingC2,
                spaceFillingMinDist,
                spaceFillingPhiP,
                spaceFillingPhiP50,
            ]
        ]
    )
)



"""
Gaussian Process Regression: Cantilever beam model using HMAT
=============================================================
"""

# %%
# In this example, we create a Gaussian Process Regression metamodel (refer to :ref:`gaussian_process_regression`)
# of the :ref:`cantilever beam <use-case-cantilever-beam>`.
# We use a squared exponential covariance kernel for the Gaussian process. In order to estimate the hyper-parameters, we use a design of experiments of size is 20.


# %%
# Definition of the model
# -----------------------

# %%
from openturns.usecases import cantilever_beam
import openturns as ot
import openturns.viewer as otv


# %%
# We load the cantilever beam use case :
cb = cantilever_beam.CantileverBeam()

# %%
# We define the function which evaluates the output depending on the inputs.
model = cb.model

# %%
# Then we define the distribution of the input random vector.
dim = cb.dim  # number of inputs
myDistribution = cb.distribution

# %%
# Create the design of experiments
# --------------------------------

# %%
# We consider a simple Monte-Carlo sample as a design of experiments. This is why we generate an input sample using the `getSample` method of the distribution.
# Then we evaluate the output using the `model` function.
sampleSize_train = 20
X_train = myDistribution.getSample(sampleSize_train)
Y_train = model(X_train)

# %%
# The following figure presents the distribution of the vertical deviations `Y` on the training sample.
# We observe that the large deviations occur less often.
histo = ot.HistogramFactory().build(Y_train).drawPDF()
histo.setXTitle("Vertical deviation (cm)")
histo.setTitle("Distribution of the vertical deviation")
histo.setLegends([""])
view = otv.View(histo)

# %%
# Create the metamodel
# --------------------

# %%
# We rely on `H-Matrix` approximation for accelerating the evaluation.
# We change default parameters (compression, recompression) to higher values. The model is less accurate but very fast to build & evaluate.
ot.ResourceMap.SetAsString("GaussianProcessFitter-LinearAlgebra", "HMAT")
ot.ResourceMap.SetAsScalar("HMatrix-AssemblyEpsilon", 1e-5)
ot.ResourceMap.SetAsScalar("HMatrix-RecompressionEpsilon", 1e-4)

# %%
# In order to create the GP metamodel, we first select a constant trend with the :class:`~openturns.ConstantBasisFactory` class.
# Then we use a squared exponential covariance kernel.
# The :class:`~openturns.SquaredExponential` kernel has one amplitude coefficient and 4 scale coefficients.
# This is because this covariance kernel is anisotropic : each of the 4 input variables is associated with its own scale coefficient.
basis = ot.ConstantBasisFactory(dim).build()
covarianceModel = ot.SquaredExponential(dim)

# %%
# Typically, the optimization algorithm is quite good at setting sensible optimization bounds.
# In this case, however, the range of the input domain is extreme.
print("Lower and upper bounds of X_train:")
print(X_train.getMin(), X_train.getMax())

# %%
# Finally, we use the :class:`~openturns.GaussianProcessRegression` class to create the GP metamodel.
# It requires a training sample, a covariance kernel and a trend basis as input arguments.
# We need to set the initial scale parameter for the optimization. The upper bound of the input domain is a sensitive choice here.
# We must not forget to actually set the optimization bounds defined above.
covarianceModel.setScale(X_train.getMax())
fitter = ot.GaussianProcessFitter(X_train, Y_train, covarianceModel, basis)
fitter.run()
fitter_result = fitter.getResult()
algo = ot.GaussianProcessRegression(fitter_result)


# %%
# The `run` method has optimized the hyperparameters of the metamodel.
#
# We can then print the constant trend of the metamodel, which have been estimated using the least squares method.
algo.run()
result = algo.getResult()
gpMetamodel = result.getMetaModel()

# %%
# The `getTrendCoefficients` method returns the coefficients of the trend.
print(result.getTrendCoefficients())

# %%
# We can also print the hyperparameters of the covariance model, which have been estimated by maximizing the likelihood.
result.getCovarianceModel()

# %%
# Validate the metamodel
# ----------------------

# %%
# We finally want to validate the Kriging metamodel. This is why we generate a validation sample with size 100 and we evaluate the output of the model on this sample.

# %%
sampleSize_test = 100
X_test = myDistribution.getSample(sampleSize_test)
Y_test = model(X_test)

# %%
# The :class:`~openturns.MetaModelValidation` class is designed to validate the surrogate models.
# To create it, we use a validation sample and a metamodel.
metamodelPredictions = gpMetamodel(X_test)
val = ot.MetaModelValidation(Y_test, metamodelPredictions)

# %%
# The :meth:`~openturns.MetaModelValidation.computeR2Score` computes the R2 coefficient of determination.
r2Score = val.computeR2Score()[0]
print("R2=", r2Score)

# %%
# The residuals are the difference between the model and the metamodel.
residualsSample = val.getResidualSample()
graph = ot.HistogramFactory().build(residualsSample).drawPDF()
graph.setXTitle("Residuals (cm)")
graph.setTitle("Distribution of the residuals")
graph.setLegends([""])
view = otv.View(graph)

# %%
# We observe that the negative residuals occur with nearly the same frequency of the positive residuals: this is a first sign of good quality.

# %%
# The `drawValidation` method allows one to compare the observed outputs and the metamodel outputs.

# %%
# sphinx_gallery_thumbnail_number = 3
graph = val.drawValidation()
graph.setTitle("R2 = %.2f%%" % (100 * r2Score))
view = otv.View(graph)

# %%
otv.View.ShowAll()



#!/usr/bin/env python

import inspect
import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott
from pathlib import Path
import re

ot.TESTPREAMBLE()

# find all rst files
rst_files = (
    Path(__file__).parents[1].joinpath("doc").joinpath("user_manual").rglob("*.rst")
)
rst_lines = []
for rst_file in rst_files:
    with open(rst_file) as f:
        rst_lines += f.read().splitlines()

# find all instantiable classes
instantiables = []
for mod in [ot, otexp, ott]:
    for name, obj in inspect.getmembers(mod):
        if inspect.isclass(obj):
            cn = obj.__name__
            if (
                cn.startswith("_")
                or cn.endswith("Collection")
                or cn.endswith("Implementation")
            ):
                continue
            try:
                instance = obj()
                instantiables.append(obj)
            except Exception:
                pass

# find missing docstrings
count_class = 0
count_class_undoc = 0
count_methods = 0
count_methods_undoc = 0

for class_ in instantiables:
    cn = class_.__name__
    count_class += 1
    found = False
    for line in rst_lines:
        if re.search(rf"{cn}", line) is not None:
            found = True
            break
    if not found:
        count_class_undoc += 1
        print(f"{cn} class")

# find all static functions
for name, mod in inspect.getmembers(ot):
    if inspect.ismodule(mod):
        modn = mod.__name__
        if not modn.startswith("openturns") or modn == "openturns":
            continue
        if "_" in modn:
            continue

        for attr_name in dir(mod):
            if "_" in attr_name:
                continue
            obj = getattr(mod, attr_name)

            if not inspect.isclass(obj) and callable(obj):
                symboln = f"{modn.split('.')[1]}.{attr_name}"
                count_methods += 1
                found = False
                for line in rst_lines:
                    if re.search(rf"\b{symboln}\b", line) is not None:
                        found = True
                        break
                if not found:
                    count_methods_undoc += 1
                    print(f"{symboln} method")

print(
    f"-- undocumented classes: {count_class_undoc} ({100.0 * count_class_undoc / count_class:.2f}%) --"
)
print(
    f"-- undocumented methods: {count_methods_undoc} ({100.0 * count_methods_undoc / count_methods:.2f}%) --"
)
if count_class_undoc + count_methods_undoc > 15:
    raise ValueError(
        f"too much undocumented class/methods {count_class_undoc + count_methods_undoc}"
    )



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Normal(4)
size = 10
experiment = ot.MonteCarloExperiment(distribution, size)
print("experiment = ", experiment)
sample, weights = experiment.generateWithWeights()
print("sample = ", repr(sample))
print("weights = ", repr(weights))



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import openturns.experimental as otexp


def build_gpr_result(inputSample, outputSample, covarianceModel, basis):
    """
    From data & covariance model, build a Gaussian Process Regression result
    """
    fitter_algo = ot.GaussianProcessFitter(
        inputSample, outputSample, covarianceModel, basis
    )
    fitter_algo.setOptimizeParameters(False)  # do not optimize hyper-parameters
    fitter_algo.run()
    fitter_result = fitter_algo.getResult()
    gpr_algo = ot.GaussianProcessRegression(fitter_result)
    gpr_algo.run()
    gpr_result = gpr_algo.getResult()
    return gpr_result


ot.TESTPREAMBLE()

ot.PlatformInfo.SetNumericalPrecision(3)
# GP use case
inputDimension = 2

# Learning data
levels = [8.0, 5.0]
box = ot.Box(levels)
inputSample = box.generate()
# Scale each direction
inputSample *= 10


model = ot.SymbolicFunction(["x", "y"], ["cos(0.5*x) + sin(y)"])
outputSample = model(inputSample)

# Validation data
sampleSize = 10
inputValidSample = ot.JointDistribution(2 * [ot.Uniform(0, 10.0)]).getSample(sampleSize)
outputValidSample = model(inputValidSample)

# 2) Definition of exponential model
# The parameters have been calibrated using TNC optimization
# and AbsoluteExponential models
covarianceModel = ot.SquaredExponential([7.63, 2.11], [7.38])

# 3) Basis definition
basis = ot.ConstantBasisFactory(inputDimension).build()

gpr_result = build_gpr_result(inputSample, outputSample, covarianceModel, basis)

# Define a 2D mesh
vertices = [[1.0, 0.0], [2.0, 0.0], [2.0, 1.0], [1.0, 1.0], [1.5, 0.5]]
simplicies = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4]]
mesh2D = ot.Mesh(vertices, simplicies)

# update the vertices
vertices = ot.Sample(inputSample)
vertices.add(ot.JointDistribution([ot.Uniform(0.0, 10.0)] * 2).getSample(100))

ot.RandomGenerator.SetSeed(0)
process = otexp.ConditionedGaussianProcess(gpr_result, mesh2D)

# Get a realization of the process
realization = process.getRealization()
print("realization = ", repr(realization))

# Get a sample & compare it to expectation
sample = process.getSample(5000)
mean = sample.computeMean()
print("Mean over 5000 realizations = ", repr(mean))

# Check if one can sample the process over a mesh containing conditioning points
# and 100 new points
process = otexp.ConditionedGaussianProcess(gpr_result, ot.Mesh(vertices))
realization = process.getRealization()
num = 0.0
den = 0.0
for i in range(len(inputSample)):
    num += (realization.getValueAtIndex(i) - outputSample[i]).norm()
    den += outputSample[i].norm()
error = num / den
ott.assert_almost_equal(error, 0.0, 1.0e-6, 1.0e-6)

# 2D use case - #2769
model = ot.SymbolicFunction(["x", "y"], ["cos(x) + sin(y)", "cos(0.5*x) + sin(y)"])
inputSample = ot.Sample([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0], [0.5, 0.5]])
outputSample = model(inputSample)
covarianceModel = ot.TensorizedCovarianceModel([ot.SquaredExponential([1.0, 1.0])] * 2)
vertices = [[0.3, 0.6], [0.4, 0.8]]
mesh2D = ot.Mesh(vertices)

fitter_algo = ot.GaussianProcessFitter(inputSample, outputSample, covarianceModel)
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
process = otexp.ConditionedGaussianProcess(gpr_result, mesh2D)
sample = process.getSample(3)
ott.assert_almost_equal(sample.getSize(), 3, 0, 0)
ott.assert_almost_equal(sample.getDimension(), 2, 0, 0)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(2)

for distribution in [ot.Gamma(1.5, 2.5, -0.5), ot.Gamma(15.0, 2.5)]:
    print("Distribution ", distribution)

    # Is this distribution elliptical ?
    print("Elliptical = ", distribution.isElliptical())

    # Is this distribution continuous ?
    print("Continuous = ", distribution.isContinuous())

    # Test for realization of distribution
    oneRealization = distribution.getRealization()
    print("oneRealization=", oneRealization)

    # Define a point
    point = ot.Point(distribution.getDimension(), 1.0)
    print("Point= ", point)

    # Show PDF and CDF of point
    eps = 1e-5
    DDF = distribution.computeDDF(point)
    print("ddf     =", DDF)

    LPDF = distribution.computeLogPDF(point)
    print("log pdf= %.6g" % LPDF)
    PDF = distribution.computePDF(point)
    print("pdf     =%.6g" % PDF)

    CDF = distribution.computeCDF(point)
    print("cdf= %.6g" % CDF)
    CCDF = distribution.computeComplementaryCDF(point)
    print("ccdf= %.6g" % CCDF)
    Survival = distribution.computeSurvivalFunction(point)
    print("survival= %.6g" % Survival)
    CF = distribution.computeCharacteristicFunction(point[0])
    print("characteristic function=(%.6g, %.6g)" % (CF.real, CF.imag))
    LCF = distribution.computeLogCharacteristicFunction(point[0])
    print("log characteristic function=(%.6g, %.6g)" % (LCF.real, LCF.imag))
    PDFgr = distribution.computePDFGradient(point)
    print("pdf gradient     =", PDFgr)

    CDFgr = distribution.computeCDFGradient(point)
    print("cdf gradient     =", CDFgr)

    quantile = distribution.computeQuantile(0.95)
    print("quantile=", quantile)
    print("cdf(quantile)=", distribution.computeCDF(quantile))
    # Get 95% survival function
    inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
    print("InverseSurvival=", repr(inverseSurvival))
    print(
        "Survival(inverseSurvival)=%.6f"
        % distribution.computeSurvivalFunction(inverseSurvival)
    )
    print("entropy=%.6f" % distribution.computeEntropy())

    # Confidence regions
    (
        interval,
        threshold,
    ) = distribution.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
    if distribution.getK() == 1.5:
        ott.assert_almost_equal(interval.getLowerBound(), [-0.49937], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [1.06337], 1e-4, 0.0)
    elif distribution.getK() == 15.0:
        ott.assert_almost_equal(interval.getLowerBound(), [3.1431], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [9.09027], 1e-4, 0.0)
    print("threshold=", ot.Point(1, threshold))
    levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
    print("Minimum volume level set=", levelSet)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
    print("Bilateral confidence interval=", interval)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, False
    )
    print("Unilateral confidence interval (lower tail)=", interval)
    print("beta=", ot.Point(1, beta))
    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, True
    )
    print("Unilateral confidence interval (upper tail)=", interval)
    print("beta=", ot.Point(1, beta))

    mean = distribution.getMean()
    print("mean=", mean)
    covariance = distribution.getCovariance()
    print("covariance=", covariance)
    correlation = distribution.getCorrelation()
    print("correlation=", correlation)
    spearman = distribution.getSpearmanCorrelation()
    print("spearman=", spearman)
    kendall = distribution.getKendallTau()
    print("kendall=", kendall)
    parameters = distribution.getParametersCollection()
    print("parameters=", parameters)
    print("Standard representative=", distribution.getStandardRepresentative())

    standardDeviation = distribution.getStandardDeviation()
    print("standard deviation=", standardDeviation)
    skewness = distribution.getSkewness()
    print("skewness=", skewness)
    kurtosis = distribution.getKurtosis()
    print("kurtosis=", kurtosis)

    ot.Log.Show(ot.Log.TRACE)
    validation = ott.DistributionValidation(distribution)
    validation.run()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


distribution = ot.Geometric(0.7)
size = 10000
sample = distribution.getSample(size)
factory = ot.GeometricFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedGeometric = factory.buildAsGeometric(sample)
print("Geometric          =", distribution)
print("Estimated geometric=", estimatedGeometric)
estimatedGeometric = factory.buildAsGeometric()
print("Default geometric=", estimatedGeometric)
estimatedGeometric = factory.buildAsGeometric(distribution.getParameter())
print("Geometric from parameters=", estimatedGeometric)



"""
Create a functional basis process
=================================
"""

# %%
# The objective of this example is to define
# :math:`X: \Omega \times \mathcal{D} \rightarrow \mathbb{R}^d` a multivariate stochastic
# process of dimension :math:`d` where :math:`\mathcal{D} \in \mathbb{R}^n`, as a linear
# combination of :math:`K` deterministic functions
# :math:`(\phi_i)_{i=1,\dots,K}: \mathbb{R}^n \rightarrow \mathbb{R}^d`:
#
# .. math::
#     \begin{aligned}
#       X(\omega,\vect{t})=\sum_{i=1}^KA_i(\omega)\phi_i(\vect{t})
#     \end{aligned}
#
# where :math:`\vect{A}=(A_1,\dots, A_K)` is a random vector of dimension :math:`K`.
#
# We suppose that :math:`\mathcal{M}` is discretized on the mesh :math:`\mathcal{M}` which has :math:`N`
# vertices.
#
# A realization of :math:`X` on :math:`\mathcal{M}` consists in generating a realization
# :math:`\vect{\alpha}` of the random vector :math:`\vect{A}` and in evaluating the
# functions :math:`(\phi_i)_{i=1,\dots,K}` on the mesh :math:`\mathcal{M}`.
#
# If we note
# :math:`(\vect{x}_0, \dots, \vect{x}_{N-1})` the realization of :math:`X`, where
# :math:`X(\omega, \vect{t}_k) = \vect{x}_k`, we have:
#
# .. math::
#    \begin{aligned}
#      \forall k \in [0, N-1], \quad \vect{x}_k = \sum_{i=1}^K\alpha_i\phi_i(\vect{t}_k)
#    \end{aligned}
#

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Define the coefficients distribution
mu = [2.0] * 2
sigma = [5.0] * 2
R = ot.CorrelationMatrix(2)
coefDist = ot.Normal(mu, sigma, R)

# %%
# Create a basis of functions
phi_1 = ot.SymbolicFunction(["t"], ["sin(t)"])
phi_2 = ot.SymbolicFunction(["t"], ["cos(t)^2"])
myBasis = ot.Basis([phi_1, phi_2])

# %%
# Create the mesh
myMesh = ot.RegularGrid(0.0, 0.1, 100)

# %%
# Create the process
process = ot.FunctionalBasisProcess(coefDist, myBasis, myMesh)

# %%
# Draw a sample
N = 6
sample = process.getSample(N)
graph = sample.drawMarginal(0)
graph.setTitle(str(N) + " realizations of functional basis process")
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.ResourceMap.SetAsUnsignedInteger(
    "CompoundDistribution-MarginalIntegrationNodesNumber", 24
)

observationsSize = 5
# Create a collection of distribution
conditionedDistribution = ot.Normal()
conditioningDistributionCollection = []
# First conditioning distribution: continuous/continuous
atoms = [ot.Uniform(0.0, 1.0), ot.Uniform(1.0, 2.0)]
conditioningDistributionCollection.append(ot.JointDistribution(atoms))
# Second conditioning distribution: discrete/continuous
atoms = [ot.Binomial(3, 0.5), ot.Uniform(1.0, 2.0)]
# conditioningDistributionCollection.append(ot.JointDistribution(atoms))
# Third conditioning distribution: dirac/continuous
atoms = [ot.Dirac(0.0), ot.Uniform(1.0, 2.0)]
# conditioningDistributionCollection.append(ot.JointDistribution(atoms))


for conditioning in conditioningDistributionCollection:
    print("\n" + "=" * 50 + "\n")
    print("conditioning distribution=", conditioning)
    observationsDistribution = ot.Distribution(conditionedDistribution)
    observationsDistribution.setParameter(conditioning.getMean())
    observations = observationsDistribution.getSample(observationsSize)
    distribution = ot.PosteriorDistribution(
        ot.CompoundDistribution(conditionedDistribution, conditioning),
        observations,
    )
    dim = distribution.getDimension()
    print("Distribution ", distribution)
    print("range=", distribution.getRange())
    mean = distribution.getMean()
    print("Mean ", mean)
    print("Covariance ", distribution.getCovariance())
    # Is this distribution an elliptical distribution?
    print("Elliptical distribution= ", distribution.isElliptical())

    # Has this distribution an elliptical copula?
    print("Elliptical copula= ", distribution.hasEllipticalCopula())

    # Has this distribution an independent copula?
    print("Independent copula= ", distribution.hasIndependentCopula())

    # Show PDF and CDF of mean point
    meanPDF = distribution.computePDF(mean)
    meanCDF = distribution.computeCDF(mean)
    print("Mean point= ", mean, " pdf=%.5g" % meanPDF, " cdf=%.5g" % meanCDF)
    # Get 95% quantile
    quantile = distribution.computeQuantile(0.95)
    print("Quantile=", quantile)
    print("CDF(quantile)=%.6g" % distribution.computeCDF(quantile))
    # Extract the marginals
    for i in range(dim):
        margin = distribution.getMarginal(i)
        print("margin=", margin)
        print("margin PDF=", ot.Point(1, margin.computePDF(mean[i])))
        print("margin CDF=", ot.Point(1, margin.computeCDF(mean[i])))
        print("margin quantile=", margin.computeQuantile(0.95))

    ot.Log.Show(ot.Log.TRACE)
    validation = ott.DistributionValidation(distribution)
    validation.setMomentsSamplingSize(10000)
    validation.skipCorrelation()  # slow
    validation.skipDependenceMeasures()  # slow
    validation.skipParameters()  # not yet implemented
    validation.skipGradient()  # not yet implemented
    validation.run()

# Test the constructors
conditioned = ot.Normal()
link = ot.SymbolicFunction(["y"], ["0.0", "y"])
conditioning = ot.Uniform(0.5, 1.5)
observations = conditioned.getSample(10)

# First constructor
posterior = ot.PosteriorDistribution()

# Second constructor
posterior = ot.PosteriorDistribution(
    ot.CompoundDistribution(conditioned, conditioning, link), observations
)

# Third constructor
posterior = ot.PosteriorDistribution(conditioned, ot.IndependentCopula(2), observations)

# Fourth constructor
posterior = ot.PosteriorDistribution(conditioned, conditioning, link, observations)

# Accessors to the arguments of the constructor
print("conditioned=", posterior.getConditionedDistribution())
print("conditioning=", posterior.getConditioningDistribution())
print("link=", posterior.getLinkFunction())



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


inputVars = ot.Description(["t"])
formula = ot.Description(["sin(t)", "cos(t)"])
myFunc = ot.SymbolicFunction(inputVars, formula)
tg = ot.RegularGrid(0.0, 0.1, 11)
myTrendFunc = ot.TrendTransform(myFunc, tg)

print("myTrendFunc=", myTrendFunc)
# Get the input description and dimension
print("myTrendFunc input description=", myTrendFunc.getInputDescription())
print("myTrendFunc input dimension=", myTrendFunc.getInputDimension())
# Get the output description and dimension
print("myTrendFunc output description=", myTrendFunc.getOutputDescription())
print("myTrendFunc output dimension=", myTrendFunc.getOutputDimension())
# Create a TimeSeries
data = ot.Sample(tg.getN(), formula.getSize())
for i in range(data.getSize()):
    for j in range(data.getDimension()):
        data[i, j] = i * data.getDimension() + j
ts = ot.TimeSeries(tg, data)
print("input time series =  ")
print(ts)
print("output time series = ")
print(myTrendFunc(ts))
# Get the number of calls
print("called ", myTrendFunc.getCallsNumber(), " times")



"""
Create a Monte Carlo design of experiments
==========================================
"""

# %%

# %%
# In this example we are going to create a MonteCarlo probabilistic design experiment.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Define underlying distribution, size
distribution = ot.Normal(2)
size = 50

# %%
# Create the design
experiment = ot.MonteCarloExperiment(distribution, size)
sample = experiment.generate()

# %%
# Plot the design
graph = ot.Graph("MC design", "x1", "x2")
cloud = ot.Cloud(sample, "blue", "fsquare", "")
graph.add(cloud)
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott

ratioOfU = otexp.RatioOfUniforms()
print("ratioOfU is initialized?", ratioOfU.isInitialized())
coll = ratioOfU.initialize()
print("ratioOfU is initialized?", ratioOfU.isInitialized())
print("Multistarts=", coll)
ratioOfU.setCandidateNumber(10)
print("Candidate number=", ratioOfU.getCandidateNumber())
ratioOfU.setOptimizationAlgorithm(ot.TNC())
print("Algo optim=", ratioOfU.getOptimizationAlgorithm())

# We create a distribution
# distribution = ot.Dirichlet([1, 2, 3, 4, 5, 6, 7])
R = ot.CorrelationMatrix(
    4, [1.0, 0.5, 0.2, 0.1, 0.5, 1.0, 0.5, 0.2, 0.2, 0.5, 1.0, 0.5, 0.1, 0.2, 0.5, 1.0]
)
distribution = ot.Normal([0.0] * 4, [1.0] * 4, R)
print("distribution = ", repr(distribution))


# Now sample it using RoU
ratioOfU = otexp.RatioOfUniforms(distribution)
print("ratioOfU=", repr(ratioOfU))
print("ratioOfU=", str(ratioOfU))
print("ratioOfU is initialized?", ratioOfU.isInitialized())
print("r=", ratioOfU.getR())
print("Acceptance ratio=%.4g" % ratioOfU.getAcceptanceRatio())
print("c=", ratioOfU.getC())
ratioOfU.setR(2.0)
print("r=", ratioOfU.getR())
print("Acceptance ratio=%.4g" % ratioOfU.getAcceptanceRatio())
print("c=", ratioOfU.getC())
print("One realization", ratioOfU.getRealization())
print("One sample", ratioOfU.getSample(10))
sample, tryNumber = ratioOfU.getSampleWithTryNumber(10)
print("One sample with try number", sample, "try number=", tryNumber)

# RoU based on a non-normalized logPDF
tri = ot.Triangular(0.0, 1.0, 1.0)
ot.RandomGenerator.SetSeed(1234)
logPDF = ot.SymbolicFunction("x", "log(x)")
ratioOfU = otexp.RatioOfUniforms(logPDF, ot.Interval(0, 1))
print("Acceptance ratio=%.4g" % ratioOfU.getAcceptanceRatio())
print("c=", ratioOfU.getC())
sample1 = ratioOfU.getSample(10)
ot.RandomGenerator.SetSeed(1234)
ratioOfU = otexp.RatioOfUniforms(tri)
print("ratioOfU=", ratioOfU)
sample2 = ratioOfU.getSample(10)
ott.assert_almost_equal(sample1, sample2)



import openturns as ot
import openturns.viewer as otv


domain = ot.Interval([-2.0, -1.0, 1.0], [2.0, 1.0, 3.0])
uniform = ot.Uniform(-3.0, 4.0)
dist = ot.JointDistribution([uniform] * 3)
sample = dist.getSample(200)
grid = ot.VisualTest.DrawInsideOutside(domain, sample)

otv.View(grid, figure_kw={"figsize": (4.5, 4.5)})



#!/usr/bin/env python

import openturns as ot

N = 10

splitters = [ot.LeaveOneOutSplitter(N)]
splitters.append(ot.KFoldSplitter(N, 3))
splitters.append(ot.KFoldSplitter(N, 5))
kfold = ot.KFoldSplitter(N, 5)
kfold.setRandomize(True)
splitters.append(kfold)
splitters.append(ot.KFoldSplitter(N, N))

for splitter in splitters:
    print(f"splitter={splitter}")
    size = 0
    for indices1, indices2 in splitter:
        print(f"indices={indices1}, {indices2}")

        assert len(indices1) + len(indices2) == N, "not enough indices"

        assert indices1.check(N), "dup or too large indices1"
        assert indices2.check(N), "dup or too large indices2"

        for i1 in indices1:
            assert i1 not in indices2, "indices1 in indices2"
        for i2 in indices2:
            assert i2 not in indices1, "indices2 in indices1"
        size += 1
    assert size == splitter.getSize(), "wrong size"

# try selection
indices = [8, 0]
sample = ot.Normal().getSample(N)
sample_i = sample[indices]
print(f"sample={sample} indices={indices} sample_i={sample_i}")
assert len(sample_i) == len(indices), "wrong selection len"
assert sample_i[0] == sample[indices[0]], "wrong first item in selection"



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.ChiSquare(0.5)
size = 10000
sample = distribution.getSample(size)
factory = ot.ChiSquareFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
distribution = ot.ChiSquare(1.0)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
distribution = ot.ChiSquare(2.5)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedChiSquare = factory.buildAsChiSquare(sample)
print("ChiSquare          =", distribution)
print("Estimated chiSquare=", estimatedChiSquare)
estimatedChiSquare = factory.buildAsChiSquare()
print("Default chiSquare=", estimatedChiSquare)
estimatedChiSquare = factory.buildAsChiSquare(distribution.getParameter())
print("ChiSquare from parameters=", estimatedChiSquare)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Default dimension parameter to evaluate the model
defaultDimension = 1
inputDimension = 1

# Amplitude values
amplitude = [1.0]

# Scale values
scale = [1.0]

# Frequency values
for frequency in [0.1, 0.2]:
    # Default constructor
    myDefaultModel = ot.ExponentiallyDampedCosineModel()
    print("myDefaultModel = ", myDefaultModel)

    # Second order model with parameters
    myModel = ot.ExponentiallyDampedCosineModel(scale, amplitude, frequency)
    print("myModel = ", myModel)

    timeValueOne = 1.0
    print("covariance matrix at t = ", timeValueOne, " : ", myModel(timeValueOne))
    print(
        "covariance matrix at t = ",
        -1.0 * timeValueOne,
        " : ",
        myModel(-1.0 * timeValueOne),
    )

    # Evaluation at time higher to check the decrease of the
    # exponentiallyDampedCosine values
    timeValueHigh = 15.0
    print(
        "covariance matrix at t = ",
        timeValueHigh,
        " : ",
        myModel(timeValueHigh).__str__(),
    )

    timeGrid = ot.RegularGrid(0.0, 1.0 / 3.0, 4)
    print(
        "discretized covariance over the time grid=",
        timeGrid,
        "is=",
        myModel.discretize(timeGrid),
    )



#! /usr/bin/env python

import openturns as ot
import math as m

ot.TESTPREAMBLE()


inputVars = ot.Description(["t"])
formula = ot.Description(["sin(t)", "cos(t)"])
myFunc = ot.SymbolicFunction(inputVars, formula)
tg = ot.RegularGrid(0.0, 0.1, 11)
myInverseTrendFunc = ot.InverseTrendTransform(myFunc, tg)

print("myInverseTrendFunc=", myInverseTrendFunc)
# Get the input description and dimension
print("myInverseTrendFunc input description=", myInverseTrendFunc.getInputDescription())
print("myInverseTrendFunc input dimension=", myInverseTrendFunc.getInputDimension())
# Get the output description and dimension
print(
    "myInverseTrendFunc output description=", myInverseTrendFunc.getOutputDescription()
)
print("myInverseTrendFunc output dimension=", myInverseTrendFunc.getOutputDimension())
# Create a TimeSeries
data = ot.Sample(tg.getN(), formula.getSize())
for i in range(data.getSize()):
    t = tg.getStart() + i * tg.getStep()
    data[i, 0] = i + 1 + m.sin(t)
    data[i, 1] = i + m.cos(t)
ts = ot.TimeSeries(tg, data)
print("input time series =  ")
print(ts)
print("output time series = ")
print(myInverseTrendFunc(ts))
# Get the number of calls
print("called ", myInverseTrendFunc.getCallsNumber(), " times")



"""
Use the Smolyak quadrature
===========================
"""

# %%
# The goal of this example is to use Smolyak's quadrature.
# We create a Smolyak quadrature using a Gauss-Legendre marginal
# quadrature and use it on a benchmark problem.
# In the second part, we compute the absolute error of Smolyak's quadrature
# method and compare it with the tensorized Gauss-Legendre quadrature.
# In the third part, we plot the absolute error depending on the sample
# sample and analyze the speed of convergence.

# %%
import numpy as np
import openturns as ot
import openturns.viewer as otv
from matplotlib import pyplot as plt

# %%
# In the first example, we print the nodes and weights
# Smolyak-Legendre quadrature of level 3.

uniform = ot.GaussProductExperiment(ot.Uniform(0.0, 1.0))
collection = [uniform] * 2
level = 3
print("level = ", level)
experiment = ot.SmolyakExperiment(collection, level)
nodes, weights = experiment.generateWithWeights()
print(nodes)
print(weights)

# %%
# We see that some of the weights are nonpositive. This is a significant
# difference with tensorized Gauss quadrature.

# %%
# We now create an exponential product problem from
# [morokoff1995]_, table 1, page 221.
# This example is also used in [gerstner1998]_ page 221 to demonstrate
# the properties of Smolyak's quadrature.
# It is defined by the equation :
#
# .. math::
#
#     g(\boldsymbol{x}) = (1 + 1 / d)^d \prod_{i = 1}^d x_i^{1 / d}
#
# for any :math:`\boldsymbol{x} \in [0, 1]^d` where :math:`d` is the
# dimension of the problem.
#
# We are interested in the integral:
#
# .. math::
#
#     I = \int_{[0, 1]^d} g(\boldsymbol{x}) f(\boldsymbol{x}) d \boldsymbol{x}
#
# where :math:`f(\boldsymbol{x}) = 1` is the uniform density probability
# function in the :math:`[0, 1]^d` interval.
#
# When the dimension increases, the variance goes to zero,
# but the variation in the sense of Hardy and Krause goes to infinity.
#

dimension = 5


def g_function_py(x):
    """
    Evaluates the exponential integrand function.

    Parameters
    ----------
    x : ot.Point

    Returns
    -------
    y : ot.Point
    """
    value = (1.0 + 1.0 / dimension) ** dimension
    for i in range(dimension):
        value *= x[i] ** (1.0 / dimension)
    return [value]


g_function = ot.PythonFunction(dimension, 1, g_function_py)
interval = ot.Interval([0.0] * dimension, [1.0] * dimension)
integral = 1.0
print("Exact integral = ", integral)
name = "ExponentialProduct"

# %%
# In the next cell, we evaluate the Smolyak quadrature.
# We first create a Smolyak experiment using a Gauss-Legendre
# marginal experiment.

uniform = ot.GaussProductExperiment(ot.Uniform(0.0, 1.0))
collection = [uniform] * dimension
level = 5
print("level = ", level)
experiment = ot.SmolyakExperiment(collection, level)
nodes, weights = experiment.generateWithWeights()
print("size = ", nodes.getSize())

# %%
# Then we evaluate the function values at the nodes, and
# use the `dot` product in order to compute the weighted
# sum of function values.
g_values = g_function(nodes)
g_values_point = g_values.asPoint()
approximate_integral = g_values_point.dot(weights)
lre10 = -np.log10(abs(approximate_integral - integral) / abs(integral))
print("Approx. integral = ", approximate_integral)
print("Log-relative error in base 10= %.2f" % (lre10))

# %%
# We see that Smolyak's quadrature has produced a quite accurate
# approximation of the integral.
# With only 781 nodes in dimension 5, the approximation has more than 2 correct
# digits.

# %%
# In the next cell, we use a tensorized Gauss quadrature rule and
# compute the accuracy of the quadrature.
numberOfMarginalNodes = 5
collectionOfMarginalNumberOfNodes = [numberOfMarginalNodes] * dimension
collection = [ot.Uniform(0.0, 1.0)] * dimension
distribution = ot.JointDistribution(collection)
experiment = ot.GaussProductExperiment(distribution, collectionOfMarginalNumberOfNodes)
nodes, weights = experiment.generateWithWeights()
size = nodes.getSize()
print("size = ", nodes.getSize())
g_values = g_function(nodes)
g_values_point = g_values.asPoint()
approximate_integral = g_values_point.dot(weights)
lre10 = -np.log10(abs(approximate_integral - integral) / abs(integral))
print("Approx. integral = ", approximate_integral)
print("Log-relative error in base 10= %.2f" % (lre10))

# %%
# Using 5 nodes in each dimension leads to a total number of nodes
# equal to 3125.
# This relatively large number of nodes leads to an approximate integral
# which has more than 2 correct digits.

# %%
# We want to see how the quadrature converges to the true integral
# when the number of nodes increases and the speed of convergence.
# To do this, we create a set of helper functions which evaluate the
# quadrature rule, compute the table of the absolute error versus the number of
# nodes and plot it.
#
# The next function performs Smolyak quadrature on a function on the
# unit cube using Gauss-Legendre quadrature rule.


def smolyakQuadrature(g_function, level):
    """
    Integrate a function g on the unit cube [0, 1]^d using Smolyak quadrature.

    Uses a Gauss-Legendre quadrature rule as the marginal quadrature.

    Parameters
    ----------
    g_function : ot.Function
        The integrand, with d inputs and dimension 1 output.
    level : int
        The level of Smolyak quadrature.

    Returns
    -------
    size : int
        The number of nodes in the quadrature.
    abserr : float
        The absolute error.

    """
    dimension = g_function.getInputDimension()
    uniform = ot.GaussProductExperiment(ot.Uniform(0.0, 1.0))
    collection = [uniform] * dimension
    experiment = ot.SmolyakExperiment(collection, level)
    nodes, weights = experiment.generateWithWeights()
    size = nodes.getSize()
    g_values = g_function(nodes)
    g_values_point = g_values.asPoint()
    approximate_integral = g_values_point.dot(weights)
    abserr = abs(approximate_integral - integral)
    return [size, abserr]


# %%
# Similarly, the next function uses the Gauss-Legendre quadrature rule.


def tensorizedGaussQuadrature(g_function, numberOfMarginalNodes):
    """
    Integrate a function g on the unit cube [0, 1]^d.

    Uses a tensorized Gauss-Legendre quadrature.

    Parameters
    ----------
    g_function : ot.Function
        The integrand, with d inputs and dimension 1 output.
    level : int
        The level of Smolyak quadrature.

    Returns
    -------
    size : int
        The number of nodes in the quadrature.
    abserr : float
        The absolute error.

    """
    dimension = g_function.getInputDimension()
    collection = [ot.Uniform(0.0, 1.0)] * dimension
    distribution = ot.JointDistribution(collection)
    collectionOfMarginalNumberOfNodes = [numberOfMarginalNodes] * dimension
    experiment = ot.GaussProductExperiment(
        distribution, collectionOfMarginalNumberOfNodes
    )
    nodes, weights = experiment.generateWithWeights()
    size = nodes.getSize()
    g_values = g_function(nodes)
    g_values_point = g_values.asPoint()
    approximate_integral = g_values_point.dot(weights)
    abserr = abs(approximate_integral - integral)
    return [size, abserr]


# %%
# The following function plots the absolute error versus the
# number of nodes and returns the graph.
# Moreover, it fits a linear least squares model against the data.
# The model is ([gerstner1998]_ page 222)
#
# .. math::
#     e_{abs} = c n^{-\alpha} \epsilon_m
#
# where :math:`e_{abs}` is the absolute error, :math:`c` is a constant
# parameter representing the absolute error when the number of nodes
# is equal to 1, :math:`n` is the number of nodes, :math:`\alpha` is a constant
# parameter representing the order of convergence and :math:`\epsilon_m`
# is the multiplicative residual.
# The logarithm of the previous equation is
#
# .. math::
#     \log(e_{abs}) = \log(c) - \alpha \log(n) + \epsilon_a
#
# where :math:`\epsilon_a = \log(\epsilon_m)` is the additive residual
# in logarithmic scale.
# This model states that the curve presenting the error depending on the
# number of nodes is a line with slope :math:`\alpha` in a log-log plot.
# A method with a more negative slope is favored, since it means that the
# speed of convergence is faster.


def drawQuadrature(
    size_list, abserr_list, quadratureName="Quadrature", pointStyle="bullet"
):
    # Plot the quadrature given the list of size and absolute errors.

    # Compute least squares fit
    data_logn = ot.Sample.BuildFromPoint(np.log(size_list))
    data_loge = ot.Sample.BuildFromPoint(np.log(abserr_list))
    basis = ot.SymbolicFunction(["log_n"], ["1.0", "log_n"])
    designMatrix = basis(data_logn)
    algo = ot.LinearLeastSquares(designMatrix, data_loge)
    algo.run()
    ls_fit = algo.getResult().getMetaModel()
    logerror_fit = ls_fit(basis(data_logn))
    error_fit = np.exp(logerror_fit).flatten()
    alpha = algo.getLinear()[1, 0]
    #
    graph = ot.Graph()
    cloud = ot.Cloud(size_list, abserr_list)
    cloud.setLegend(quadratureName)
    cloud.setPointStyle(pointStyle)
    graph.add(cloud)
    curve = ot.Curve(size_list, error_fit)
    curve.setLegend(r"$\alpha$ = %.3f" % (alpha))
    graph.add(curve)
    return graph


# %%
# The next two functions plots the Smolyak and tensorized Gauss quadrature.


def drawSmolyakQuadrature(level_max):
    print("Smolyak-Gauss Quadrature")
    size_list = []
    abserr_list = []
    for level in range(1, level_max):
        size, abserr = smolyakQuadrature(g_function, level)
        print("size = %d, level = %d, ea = %.3e" % (size, level, abserr))
        size_list.append(size)
        abserr_list.append(abserr)
    graph = drawQuadrature(size_list, abserr_list, "Smolyak-Gauss", "bullet")
    return graph


def drawTensorizedGaussQuadrature(n_max):
    print("Tensorized Gauss Quadrature")
    size_list = []
    abserr_list = []
    for n in range(1, n_max):
        size, abserr = tensorizedGaussQuadrature(g_function, n)
        print("size = %d, n = %d, ea = %.3e" % (size, n, abserr))
        size_list.append(size)
        abserr_list.append(abserr)
    graph = drawQuadrature(size_list, abserr_list, "Gauss-Legendre", "plus")
    return graph


# %%
# We can finally create the graph.

level_max = 14
graph = ot.Graph("Exponential problem", "$n$", "$e_{abs}$")
curve = drawSmolyakQuadrature(level_max)
graph.add(curve)
n_max = 11
curve = drawTensorizedGaussQuadrature(n_max)
graph.add(curve)
graph.setLogScale(ot.GraphImplementation.LOGXY)
graph.setLegendCorner([1.0, 1.0])
graph.setLegendPosition("upper left")
view = otv.View(
    graph,
    figure_kw={"figsize": (5.0, 3.0)},
)
plt.tight_layout()

# %%
# Display all figures
otv.View.ShowAll()



"""
Integrate a function with Gauss-Kronrod algorithm
=================================================
"""

# %%
import openturns as ot
import numpy as np
import openturns.viewer as otv

# %%
# References :
#
# - [davis1975]_ section "2.7.1.1 "The Kronrod scheme", p. 82 and section 6.2.0 "An iterative nonadaptive scheme based on Kronrod formulas" p.321.
# - [dahlquist2008]_ p.575.

# %%
# Introduction
# ============
#
# In this example, we present the :class:`~openturns.GaussKronrod` algorithm for one dimensional integration. That is, the algorithm can approximate the integral:
#
# .. math::
#    \int_a^b f(x) dx
#
# where :math:`f:[a,b] \rightarrow \mathbb{R}^p` is a function,
# :math:`[a,b] \subset \mathbb{R}` with :math:`a\leq b` is a one dimensional interval,
# :math:`p` is the dimension of the output.
# Notice that the dimension of the input must be equal to 1, but the number of outputs can be greater than 1.
#
# Suppose that we have estimated the integral with Gaussian quadrature and :math:`m` quadrature nodes.
# If we want to improve the accuracy and use more nodes, the issue is that the new nodes do not correspond to the old ones: therefore, we cannot reuse the function evaluations.
#
# The Gauss-Kronrod algorithm improves the situation by using two different methods:
#
# - a Gaussian quadrature rule with :math:`m` nodes,
# - a Kronrod extension with :math:`2m+1` nodes.
#
# The rule :math:`(G_m,K_{2m+1})` is called a Gauss-Kronrod pair. In the Kronrod extension, the first :math:`m` nodes are equal to the nodes in Gaussian quadrature.
#
# The Gaussian quadrature rule with :math:`m` nodes is exact for polynomials of degree :math:`2m-1`.
# The Kronrod extension with :math:`2m+1` nodes is designed to be exact for polynomials of degree :math:`3m+1`.
#
# The choice of the weight function :math:`w(x)` determines the nodes.
# We consider the weight :math:`w(x)=1` and the interval :math:`[a,b]=[-1,1]` (it is straightforward to generalize this for an arbitrary interval :math:`[a,b]`).
# In this case, the new :math:`m+1` nodes of the Kronrod extension interlaces with the Gaussian nodes.
# The weights are guaranteed to be positive (an essential property for numerical stability).

# %%
# Example
# =======
# The following example is from [davis1975]_ p.325:
#
# .. math::
#    \int_0^1 f(x) dx = \int_0^1 \frac{2}{2 + \sin(10 \pi x)} dx = \frac{2}{\sqrt{3}} = 1.154700538379251529.

# %%
# We first define the function as a :class:`~openturns.SymbolicFunction`.
integrand = ot.SymbolicFunction(["x"], ["2 / (2 + sin(10 * pi_ * x))"])
integrand.setOutputDescription([r"$\frac{2}{2 + sin(10 \pi x)}$"])
graph = integrand.draw(0.0, 1.0, 200)
_ = otv.View(graph)

# %%
# We see that regular spikes and valleys will make this function difficult to integrate, because of the large curvatures implied at these points.
#
# We will later count the number of function evaluations.
# But a small amount of function evaluations has already been used for the plot and this is why we must take it into account.
before_evaluation_number = integrand.getEvaluationCallsNumber()
before_evaluation_number

# %%
# Basic use
# =========
# We first choose the Gauss-Kronrod rule. Six quadratures are available: we select the "G11K23" rule.
# It uses 11 nodes from a Gauss rule and 23 nodes from a Kronrod rule, re-using the nodes from the Gauss rule.
quadrature_rule = ot.GaussKronrodRule(ot.GaussKronrodRule.G11K23)

# %%
# We set the maximum number of sub-intervals and the maximum absolute error.
maximumSubIntervals = 100
maximumError = 1.0e-8
algo = ot.GaussKronrod(maximumSubIntervals, maximumError, quadrature_rule)
interval = ot.Interval(0.0, 1.0)
computed = algo.integrate(integrand, interval)
computed[0]

# %%
# Notice that the algorithm can integrate a function which has several outputs (but the number of inputs is restricted to 1).
# This is why we use the index `[0]` of `computed`, since :meth:`~openturns.GaussKronrod.integrate` returns a :class:`~openturns.Point`.
# In order to check this computation, we compute the log-relative error in base 10.
# In most cases (except when the exponent of the two numbers are different), this represents the number of correct digits in base 10.
expected = 1.154700538379251529
LRE_10 = -np.log10(abs(computed[0] - expected) / abs(expected))
LRE_10

# %%
# The method computes more than 14 digits correctly. Given that 17 digits is the best we can, this is an astonishing performance.
#
# We then compute the number of function evaluations.
after_evaluation_number = integrand.getEvaluationCallsNumber()
number_of_calls = after_evaluation_number - before_evaluation_number
number_of_calls

# %%
# Advanced use
# ============
#
# The Gauss-Kronrod algorithm strives to produce an approximated integral which actual error is less than the tolerance.
# The algorithm estimates the error, which may be used to guess the accuracy in the situation where the exact value is unknown (this is the general use case, of course).
#
# In order to get the error estimated by the algorithm, we use the third input argument of the :meth:`~openturns.GaussKronrod.integrate` method.
error = ot.Point()
computed = algo.integrate(integrand, interval, error)
computed[0]

# %%
# The variable `error` now contains the error estimate from the algorithm.
error[0]

# %%
# We see that the error estimate is a little lower than the tolerance, which indicates that the integral should be correctly approximated.

# %%
# In the next advanced example, let us now use the powerful feature of the
# algorithm that is, the ability to get the intermediate sub-intervals
# used by the algorithm.
error = ot.Point()
lowerBound = 0.0
upperBound = 1.0
ai = ot.Point()
bi = ot.Point()
ei = ot.Point()
fi = ot.Sample()
computed = algo.integrate(integrand, lowerBound, upperBound, error, ai, bi, fi, ei)
computed[0]

# %%
# The error still contains the estimate of the error.
error[0]

# %%
# During the algorithm, a collection of subintegrals
#
# .. math::
#    \int_{a_i}^{b_i} f(x) dx
#
# are approximated.
# The outputs :math:`a_i` and :math:`b_i` contain the subintervals used in the algorithm.
print("ai:", ai)
print("bi:", bi)

# %%s
# The corresponding value of the integrals are in :math:`f_i`.
# Since :math:`f` can be a multidimensional point, this is a :class:`~openturns.Sample`, which dimension corresponds to the output dimension of the function :math:`f`.
print("fi:", fi)

# %%
# The sum of these sub-integrals is the value of the integral:
sum(fi)[0]

# %%
# The estimated error of each integral is in :math:`e_i`:
number_of_intervals = ai.getDimension()
print("number of intervals:", number_of_intervals)
for i in range(number_of_intervals):
    print(
        f"Integral #{i} : [{ai[i]:.4f}, {bi[i]:.4f}] is {fi[i, 0]:.4f} with error {ei[i]:.3e}"
    )

otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import os

ot.TESTPREAMBLE()

f = ot.PointToFieldFunction()

# load
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager("pyp2ff.xml"))
study.load()
study.fillObject("f", f)

x = [4, 5]
y = f(x)
print(y[0])
assert y[0] == [5.0], "wrong y value"

os.remove("pyp2ff.xml")
